"""
Converters are used to convert documents between python dictionaries and Doc objects.
There are two types of converters: readers and writers. Readers convert dictionaries to
Doc objects, and writers convert Doc objects to dictionaries.
"""
import contextlib
import inspect
from copy import copy
from types import FunctionType
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Optional,
    Sequence,
    Tuple,
    Union,
)

from confit.registry import ValidatedFunction
from spacy.tokenizer import Tokenizer
from spacy.tokens import Doc, Span

import edsnlp
from edsnlp import registry
from edsnlp.core import PipelineProtocol
from edsnlp.utils.bindings import BINDING_GETTERS
from edsnlp.utils.span_getters import (
    SpanGetterArg,
    SpanSetterArg,
    get_spans,
    get_spans_with_group,
    set_spans,
)

FILENAME = "__FILENAME__"
SPAN_BUILTIN_ATTRS = ("sent", "label_", "kb_id_", "text")

SCHEMA = {}

_DEFAULT_TOKENIZER = None


def without_filename(d):
    d.pop(FILENAME, None)
    return d


def validate_kwargs(converter, kwargs):
    converter: FunctionType = copy(converter)
    spec = inspect.getfullargspec(converter)
    first = spec.args[0]
    converter.__annotations__[first] = Optional[Any]
    converter.__defaults__ = (None, *(spec.defaults or ())[-len(spec.args) + 1 :])
    vd = ValidatedFunction(converter, {"arbitrary_types_allowed": True})
    model = vd.init_model_instance(**kwargs)
    d = {
        k: v
        for k, v in model._iter()
        if (k in model.__fields__ or model.__fields__[k].default_factory)
    }
    d.pop("v__duplicate_kwargs", None)  # see pydantic ValidatedFunction code
    d.pop(vd.v_args_name, None)
    d.pop(first, None)
    return {**(d.pop(vd.v_kwargs_name, None) or {}), **d}


class SequenceStr:
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, value, config=None) -> Sequence[str]:
        if isinstance(value, str):
            return [value]
        if isinstance(value, list):
            return value
        raise ValueError("Not a string or list of strings")


if TYPE_CHECKING:
    SequenceStr = Union[str, Sequence[str]]  # noqa: F811


class AttributesMappingArg:
    """
    A mapping from JSON attributes to Span extensions (can be a list too).

    For instance:

    - `doc_attributes={"datetime": "note_datetime"}` will map the `datetime` JSON
      attribute to the `note_datetime` extension.
    - `doc_attributes="note_datetime"` will map the `note_datetime` JSON attribute to
      the `note_datetime` extension.
    - `span_attributes=["negation", "family"]` will map the `negation` and `family` JSON
      attributes to the `negation` and `family` extensions.
    """

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, value, config=None) -> Dict[str, str]:
        return validate_attributes_mapping(value)


if TYPE_CHECKING:
    AttributesMappingArg = Union[str, Sequence[str], Dict[str, str]]  # noqa: F811


def validate_attributes_mapping(value: AttributesMappingArg) -> Dict[str, str]:
    if isinstance(value, str):
        return {value: value}
    if isinstance(value, list):
        return {item: item for item in value}
    else:
        return value


def get_current_tokenizer():
    global _DEFAULT_TOKENIZER
    if _DEFAULT_TOKENIZER is None:
        _DEFAULT_TOKENIZER = edsnlp.blank("eds").tokenizer
    return _DEFAULT_TOKENIZER


@contextlib.contextmanager
def set_current_tokenizer(tokenizer):
    global _DEFAULT_TOKENIZER
    old = _DEFAULT_TOKENIZER
    if tokenizer:
        _DEFAULT_TOKENIZER = tokenizer
    yield
    _DEFAULT_TOKENIZER = old


@registry.factory.register("eds.standoff_dict2doc", spacy_compatible=False)
class StandoffDict2DocConverter:
    """
    !!! note "Why does BRAT/Standoff need a converter ?"

        You may wonder : why do I need a converter ? Since BRAT is already a NLP
        oriented format, it should be straightforward to convert it to a Doc object.

        Indeed, we do provide a default converter for the BRAT standoff format, but we
        also acknowledge that there may be more than one way to convert a standoff
        document to a Doc object. For instance, an annotated span may be used to
        represent a relation between two smaller included entities, or another entity
        scope, etc.

        In such cases, we recommend you use a custom converter as described
        [here](/data/converters/#custom-converter).

    Examples
    --------

    ```{ .python .no-check }
    # Any kind of reader (`edsnlp.data.read/from_...`) can be used here
    docs = edsnlp.data.read_standoff(
        "path/to/standoff",
        converter="standoff",  # set by default

        # Optional parameters
        tokenizer=tokenizer,
        span_setter={"ents": True, "*": True},
        span_attributes={"negation": "negated"},
        keep_raw_attribute_values=False,
        default_attributes={"negated": False, "temporality": "present"},
    )
    ```

    Parameters
    ----------
    nlp: Optional[PipelineProtocol]
        The pipeline object (optional and likely not needed, prefer to use the
        `tokenizer` directly argument instead).
    tokenizer: Optional[spacy.tokenizer.Tokenizer]
        The tokenizer instance used to tokenize the documents. Likely not needed since
        by default it uses the current context tokenizer :

        - the tokenizer of the next pipeline run by `.map_pipeline` in a
          [LazyCollection][edsnlp.core.lazy_collection.LazyCollection].
        - or the `eds` tokenizer by default.
    span_setter : SpanSetterArg
        The span setter to use when setting the spans in the documents. Defaults to
        setting the spans in the `ents` attribute, and creates a new span group for
        each JSON entity label.
    span_attributes : Optional[AttributesMappingArg]
        Mapping from BRAT attributes to Span extensions (can be a list too).
        By default, all attributes are imported as Span extensions with the same name.
    keep_raw_attribute_values : bool
        Whether to keep the raw attribute values (as strings) or to convert them to
        Python objects (e.g. booleans).
    default_attributes : AttributesMappingArg
        How to set attributes on spans for which no attribute value was found in the
        input format. This is especially useful for negation, or frequent attributes
        values (e.g. "negated" is often False, "temporal" is often "present"), that
        annotators may not want to annotate every time.
    notes_as_span_attribute : Optional[str]
        If set, the AnnotatorNote annotations will be concatenated and stored in a span
        attribute with this name.
    split_fragments : bool
        Whether to split the fragments into separate spans or not. If set to False, the
        fragments will be concatenated into a single span.
    """

    def __init__(
        self,
        nlp: Optional[PipelineProtocol] = None,
        *,
        tokenizer: Optional[Tokenizer] = None,
        span_setter: SpanSetterArg = {"ents": True, "*": True},
        span_attributes: Optional[AttributesMappingArg] = None,
        keep_raw_attribute_values: bool = False,
        bool_attributes: SequenceStr = [],
        default_attributes: AttributesMappingArg = {},
        notes_as_span_attribute: Optional[str] = None,
        split_fragments: bool = True,
    ):
        self.tokenizer = tokenizer or (nlp.tokenizer if nlp is not None else None)
        self.span_setter = span_setter
        self.span_attributes = span_attributes  # type: ignore
        self.keep_raw_attribute_values = keep_raw_attribute_values
        self.default_attributes = default_attributes
        self.notes_as_span_attribute = notes_as_span_attribute
        self.split_fragments = split_fragments
        for attr in bool_attributes:
            self.default_attributes[attr] = False

    def __call__(self, obj):
        tok = get_current_tokenizer() if self.tokenizer is None else self.tokenizer
        doc = tok(obj["text"] or "")
        doc._.note_id = obj.get("doc_id", obj.get(FILENAME))

        spans = []

        for dst in (
            *(() if self.span_attributes is None else self.span_attributes.values()),
            *self.default_attributes,
        ):
            if not Span.has_extension(dst):
                Span.set_extension(dst, default=None)

        for ent in obj.get("entities") or ():
            fragments = (
                [
                    {
                        "begin": min(f["begin"] for f in ent["fragments"]),
                        "end": max(f["end"] for f in ent["fragments"]),
                    }
                ]
                if not self.split_fragments
                else ent["fragments"]
            )
            for fragment in fragments:
                span = doc.char_span(
                    fragment["begin"],
                    fragment["end"],
                    label=ent["label"],
                    alignment_mode="expand",
                )
                if self.notes_as_span_attribute and ent["notes"]:
                    ent["attributes"][self.notes_as_span_attribute] = "|".join(
                        note["value"] for note in ent["notes"]
                    )
                for label, value in ent["attributes"].items():
                    new_name = (
                        self.span_attributes.get(label, None)
                        if self.span_attributes is not None
                        else label
                    )
                    if self.span_attributes is None and not Span.has_extension(
                        new_name
                    ):
                        Span.set_extension(new_name, default=None)

                    if new_name:
                        value = True if value is None else value
                        if not self.keep_raw_attribute_values:
                            value = (
                                True
                                if value in ("True", "true")
                                else False
                                if value in ("False", "false")
                                else value
                            )
                        span._.set(new_name, value)

                spans.append(span)

        set_spans(doc, spans, span_setter=self.span_setter)
        for attr, value in self.default_attributes.items():
            for span in spans:
                if span._.get(attr) is None:
                    span._.set(attr, value)

        return doc


@registry.factory.register("eds.standoff_doc2dict", spacy_compatible=False)
class StandoffDoc2DictConverter:
    """
    Examples
    --------

    ```{ .python .no-check }
    # Any kind of writer (`edsnlp.data.read/from_...`) can be used here
    edsnlp.data.write_standoff(
        docs,
        converter="standoff",  # set by default

        # Optional parameters
        span_getter={"ents": True},
        span_attributes={"negation": "negated"},
    )
    # or docs.to_standoff(...) if it's already a
    # [LazyCollection][edsnlp.core.lazy_collection.LazyCollection]
    ```

    Parameters
    ----------
    span_getter: SpanGetterArg
        The span getter to use when getting the spans from the documents. Defaults to
        getting the spans in the `ents` attribute.
    span_attributes: AttributesMappingArg
        Mapping from Span extensions to JSON attributes (can be a list too).
        By default, no attribute is exported, except `note_id`.
    """

    def __init__(
        self,
        *,
        span_getter: Optional[SpanGetterArg] = {"ents": True},
        span_attributes: AttributesMappingArg = {},
    ):
        self.span_getter = span_getter
        self.span_attributes = span_attributes

    def __call__(self, doc):
        spans = get_spans(doc, self.span_getter)
        obj = {
            FILENAME: doc._.note_id,
            "doc_id": doc._.note_id,
            "text": doc.text,
            "entities": [
                {
                    "entity_id": i,
                    "fragments": [
                        {
                            "begin": ent.start_char,
                            "end": ent.end_char,
                        }
                    ],
                    "attributes": {
                        obj_name: getattr(ent._, ext_name)
                        for ext_name, obj_name in self.span_attributes.items()
                        if ent._.has(ext_name)
                    },
                    "label": ent.label_,
                }
                for i, ent in enumerate(sorted(dict.fromkeys(spans)))
            ],
        }
        return obj


@registry.factory.register("eds.omop_dict2doc", spacy_compatible=False)
class OmopDict2DocConverter:
    """
    Examples
    --------

    ```{ .python .no-check }
    # Any kind of reader (`edsnlp.data.read/from_...`) can be used here
    docs = edsnlp.data.from_pandas(
        df,
        converter="omop",

        # Optional parameters
        tokenizer=tokenizer,
        doc_attributes=["note_datetime"],

        # Parameters below should only matter if you plan to import entities
        # from the dataframe. If the data doesn't contain pre-annotated
        # entities, you can ignore these.
        span_setter={"ents": True, "*": True},
        span_attributes={"negation": "negated"},
        default_attributes={"negated": False, "temporality": "present"},
    )
    ```

    Parameters
    ----------
    nlp: Optional[PipelineProtocol]
        The pipeline object (optional and likely not needed, prefer to use the
        `tokenizer` directly argument instead).
    tokenizer: Optional[spacy.tokenizer.Tokenizer]
        The tokenizer instance used to tokenize the documents. Likely not needed since
        by default it uses the current context tokenizer :

        - the tokenizer of the next pipeline run by `.map_pipeline` in a
          [LazyCollection][edsnlp.core.lazy_collection.LazyCollection].
        - or the `eds` tokenizer by default.
    span_setter: SpanSetterArg
        The span setter to use when setting the spans in the documents. Defaults to
        setting the spans in the `ents` attribute, and creates a new span group for
        each JSON entity label.
    doc_attributes: AttributesMappingArg
        Mapping from JSON attributes to additional Span extensions (can be a list too).
        By default, all attributes are imported as Doc extensions with the same name.
    span_attributes: Optional[AttributesMappingArg]
        Mapping from JSON attributes to Span extensions (can be a list too).
        By default, all attributes are imported as Span extensions with the same name.
    default_attributes: AttributesMappingArg
        How to set attributes on spans for which no attribute value was found in the
        input format. This is especially useful for negation, or frequent attributes
        values (e.g. "negated" is often False, "temporal" is often "present"), that
        annotators may not want to annotate every time.
    """

    def __init__(
        self,
        nlp: Optional[PipelineProtocol] = None,
        *,
        tokenizer: Optional[PipelineProtocol] = None,
        span_setter: SpanSetterArg = {"ents": True, "*": True},
        doc_attributes: AttributesMappingArg = {},
        span_attributes: Optional[AttributesMappingArg] = None,
        default_attributes: AttributesMappingArg = {},
        bool_attributes: SequenceStr = [],
    ):
        self.tokenizer = tokenizer or (nlp.tokenizer if nlp is not None else None)
        self.span_setter = span_setter
        self.doc_attributes = doc_attributes
        self.span_attributes = span_attributes
        self.default_attributes = default_attributes
        for attr in bool_attributes:
            self.default_attributes[attr] = False

    def __call__(self, obj):
        tok = get_current_tokenizer() if self.tokenizer is None else self.tokenizer
        doc = tok(obj["note_text"] or "")
        doc._.note_id = obj.get("note_id", obj.get(FILENAME))
        for obj_name, ext_name in self.doc_attributes.items():
            if not Doc.has_extension(ext_name):
                Doc.set_extension(ext_name, default=None)
            doc._.set(ext_name, obj.get(obj_name))

        spans = []

        for dst in (
            *(() if self.span_attributes is None else self.span_attributes.values()),
            *self.default_attributes,
        ):
            if not Span.has_extension(dst):
                Span.set_extension(dst, default=None)

        for ent in obj.get("entities") or ():
            ent = dict(ent)
            span = doc.char_span(
                ent.pop("start_char"),
                ent.pop("end_char"),
                label=ent.pop("note_nlp_source_value"),
                alignment_mode="expand",
            )
            for label, value in ent.items():
                new_name = (
                    self.span_attributes.get(label, None)
                    if self.span_attributes is not None
                    else label
                )
                if self.span_attributes is None and not Span.has_extension(new_name):
                    Span.set_extension(new_name, default=None)

                if new_name:
                    span._.set(new_name, value)
            spans.append(span)

        set_spans(doc, spans, span_setter=self.span_setter)
        for attr, value in self.default_attributes.items():
            for span in spans:
                if span._.get(attr) is None:
                    span._.set(attr, value)
        return doc


@registry.factory.register("eds.omop_doc2dict", spacy_compatible=False)
class OmopDoc2DictConverter:
    """
    Examples
    --------

    ```{ .python .no-check }
    # Any kind of writer (`edsnlp.data.write/to_...`) can be used here
    df = edsnlp.data.to_pandas(
        docs,
        converter="omop",

        # Optional parameters
        span_getter={"ents": True},
        doc_attributes=["note_datetime"],
        span_attributes=["negation", "family"],
    )
    # or docs.to_pandas(...) if it's already a
    # [LazyCollection][edsnlp.core.lazy_collection.LazyCollection]
    ```

    Parameters
    ----------
    span_getter: SpanGetterArg
        The span getter to use when getting the spans from the documents. Defaults to
        getting the spans in the `ents` attribute.
    doc_attributes: AttributesMappingArg
        Mapping from Doc extensions to JSON attributes (can be a list too).
        By default, no doc attribute is exported, except `note_id`.
    span_attributes: AttributesMappingArg
        Mapping from Span extensions to JSON attributes (can be a list too).
        By default, no attribute is exported.
    """

    def __init__(
        self,
        *,
        span_getter: SpanGetterArg = {"ents": True},
        doc_attributes: AttributesMappingArg = {},
        span_attributes: AttributesMappingArg = {},
    ):
        self.span_getter = span_getter
        self.doc_attributes = doc_attributes
        self.span_attributes = span_attributes

    def __call__(self, doc):
        spans = get_spans(doc, self.span_getter)
        span_binding_getters = {
            obj_name: BINDING_GETTERS[
                ("_." + ext_name)
                if ext_name.split(".")[0] not in SPAN_BUILTIN_ATTRS
                else ext_name
            ]
            for ext_name, obj_name in self.span_attributes.items()
        }
        obj = {
            FILENAME: doc._.note_id,
            "note_id": doc._.note_id,
            "note_text": doc.text,
            **{
                obj_name: getattr(doc._, ext_name)
                for ext_name, obj_name in self.doc_attributes.items()
                if doc._.has(ext_name)
            },
            "entities": [
                {
                    "note_nlp_id": i,
                    "start_char": ent.start_char,
                    "end_char": ent.end_char,
                    "lexical_variant": ent.text,
                    "note_nlp_source_value": ent.label_,
                    **{
                        obj_name: getter(ent)
                        for obj_name, getter in span_binding_getters.items()
                    },
                }
                for i, ent in enumerate(sorted(dict.fromkeys(spans)))
            ],
        }
        return obj


@registry.factory.register("eds.ents_doc2dict", spacy_compatible=False)
class EntsDoc2DictConverter:
    """
    Parameters
    ----------
    span_getter: SpanGetterArg
        The span getter to use when getting the spans from the documents. Defaults to
        getting the spans in the `ents` attribute.
    doc_attributes: AttributesMappingArg
        Mapping from Doc extensions to JSON attributes (can be a list too).
        By default, no doc attribute is exported, except `note_id`.
    span_attributes: AttributesMappingArg
        Mapping from Span extensions to JSON attributes (can be a list too).
        By default, no attribute is exported.
    """

    def __init__(
        self,
        *,
        span_getter: SpanGetterArg = {"ents": True},
        doc_attributes: AttributesMappingArg = {},
        span_attributes: AttributesMappingArg = {},
    ):
        self.span_getter = span_getter
        self.doc_attributes = doc_attributes
        self.span_attributes = span_attributes

    def __call__(self, doc):
        span_binding_getters = {
            obj_name: BINDING_GETTERS[
                ("_." + ext_name)
                if ext_name.split(".")[0] not in SPAN_BUILTIN_ATTRS
                else ext_name
            ]
            for ext_name, obj_name in self.span_attributes.items()
        }
        doc_attributes_values = {
            obj_name: BINDING_GETTERS["_." + ext_name](doc)
            for ext_name, obj_name in self.doc_attributes.items()
        }
        return [
            {
                "note_id": doc._.note_id,
                "start": ent.start_char,
                "end": ent.end_char,
                "label": ent.label_,
                "lexical_variant": ent.text,
                "span_type": group,  # for backward compatibility
                **{
                    obj_name: getter(ent)
                    for obj_name, getter in span_binding_getters.items()
                },
                **doc_attributes_values,
            }
            for ent, group in sorted(
                dict(get_spans_with_group(doc, self.span_getter)).items()
            )
        ]


def get_dict2doc_converter(
    converter: Union[str, Callable], kwargs
) -> Tuple[Callable, Dict]:
    kwargs_to_init = False
    if not callable(converter):
        available = edsnlp.registry.factory.get_available()
        try:
            filtered = [
                name
                for name in available
                if converter == name or (converter in name and "dict2doc" in name)
            ]
            converter = edsnlp.registry.factory.get(filtered[0])
            converter = converter(**kwargs).instantiate(nlp=None)
            kwargs = {}
            return converter, kwargs
        except (KeyError, IndexError):
            available = [v for v in available if "dict2doc" in v]
            raise ValueError(
                f"Cannot find converter for format {converter}. "
                f"Available converters are {', '.join(available)}"
            )
    if isinstance(converter, type) or kwargs_to_init:
        return converter(**kwargs), {}
    return converter, validate_kwargs(converter, kwargs)


def get_doc2dict_converter(
    converter: Union[str, Callable], kwargs
) -> Tuple[Callable, Dict]:
    if not callable(converter):
        available = edsnlp.registry.factory.get_available()
        try:
            filtered = [
                name
                for name in available
                if converter == name or (converter in name and "doc2dict" in name)
            ]
            converter = edsnlp.registry.factory.get(filtered[0])
            converter = converter(**kwargs).instantiate(nlp=None)
            kwargs = {}
            return converter, kwargs
        except (KeyError, IndexError):
            available = [v for v in available if "doc2dict" in v]
            raise ValueError(
                f"Cannot find converter for format {converter}. "
                f"Available converters are {', '.join(available)}"
            )
    return converter, validate_kwargs(converter, kwargs)



"""`eds.aids` pipeline"""
import itertools
from typing import Any, Dict, List, Optional, Union

from edsnlp.core import PipelineProtocol
from edsnlp.pipes.base import SpanSetterArg
from edsnlp.pipes.ner.disorders.base import DisorderMatcher
from edsnlp.pipes.qualifiers.hypothesis import HypothesisQualifier
from edsnlp.pipes.qualifiers.hypothesis.factory import (
    DEFAULT_CONFIG as DEFAULT_CONFIG_HYP,
)
from edsnlp.pipes.qualifiers.negation.factory import (
    DEFAULT_CONFIG as DEFAULT_CONFIG_NEG,
)
from edsnlp.pipes.qualifiers.negation.negation import NegationQualifier

from .patterns import default_patterns


class AIDSMatcher(DisorderMatcher):
    """
    The `eds.aids` pipeline component extracts mentions of AIDS. It will notably match:

    - Mentions of VIH/HIV at the SIDA/AIDS stage
    - Mentions of VIH/HIV with opportunistic(s) infection(s)

    ??? info "Details of the used patterns"
        ```{ .python .no-check }
        # fmt: off
        --8<-- "edsnlp/pipes/ner/disorders/AIDS/patterns.py"
        # fmt: on
        ```

    !!! warning "On HIV infection"

        pre-AIDS HIV infection are not extracted, only AIDS.

    Extensions
    ----------
    On each span `span` that match, the following attributes are available:

    - `span._.detailed_status`: set to `"PRESENT"`
    - `span._.assigned`: dictionary with the following keys, if relevant:
        - `opportunist`: list of opportunist infections extracted around the HIV mention
        - `stage`: stage of the HIV infection

    Examples
    --------
    ```python
    import edsnlp, edsnlp.pipes as eds

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(eds.sentences())
    nlp.add_pipe(
        eds.normalizer(
            accents=True,
            lowercase=True,
            quotes=True,
            spaces=True,
            pollution=dict(
                information=True,
                bars=True,
                biology=True,
                doctors=True,
                web=True,
                coding=True,
                footer=True,
            ),
        ),
    )
    nlp.add_pipe(f"eds.aids")
    ```

    Below are a few examples:

    --8<-- "docs/assets/fragments/aids-examples.md"

    Parameters
    ----------
    nlp : Optional[PipelineProtocol]
        The pipeline object
    name : Optional[str]
        The name of the component
    patterns: Union[Dict[str, Any], List[Dict[str, Any]]]
        The patterns to use for matching
    label : str
        The label to use for the `Span` object and the extension
    span_setter : SpanSetterArg
        How to set matches on the doc

    Authors and citation
    --------------------
    The `eds.aids` component was developed by AP-HP's Data Science team with a team of
    medical experts. A paper describing in details the development of those components
    is being drafted and will soon be available.
    """

    def __init__(
        self,
        nlp: Optional[PipelineProtocol],
        name: str = "aids",
        *,
        patterns: Union[Dict[str, Any], List[Dict[str, Any]]] = default_patterns,
        label: str = "aids",
        span_setter: SpanSetterArg = {"ents": True, "aids": True},
    ):
        super().__init__(
            nlp=nlp,
            name=name,
            label=label,
            patterns=patterns,
            include_assigned=False,
            span_setter=span_setter,
        )

        self.inner_negation = NegationQualifier(
            nlp,
            **{
                **DEFAULT_CONFIG_NEG,
                "on_ents_only": "AIDS_opportunist",
            },
        )

        self.inner_hypothesis = HypothesisQualifier(
            nlp,
            **{
                **DEFAULT_CONFIG_HYP,
                "on_ents_only": "AIDS_opportunist",
            },
        )

    def process(self, doc):
        spans = list(super().process(doc))

        doc.spans["AIDS_opportunist"] = list(
            itertools.chain.from_iterable(
                [span._.assigned.get("opportunist", []) for span in spans]
            )
        )

        doc = self.inner_negation(self.inner_hypothesis(doc))

        for span in spans:
            opportunists = span._.assigned.get("opportunist", [])
            if opportunists:
                opportunists = [
                    ent
                    for ent in opportunists
                    if not (ent._.negation or ent._.hypothesis)
                ]
            stage = "stage" in span._.assigned

            if span._.source == "hiv" and not (opportunists or stage):
                continue

            yield span

        del doc.spans["AIDS_opportunist"]



from typing import List, Optional

from edsnlp.core import PipelineProtocol, registry

from .sentences import SentenceSegmenter

DEFAULT_CONFIG = dict(
    punct_chars=None,
    ignore_excluded=True,
    use_endlines=None,
)


@registry.factory.register(
    "eds.sentences",
    assigns=["token.is_sent_start"],
    deprecated=["sentences"],
)
def create_component(
    nlp: PipelineProtocol,
    name: str = "sentences",
    *,
    punct_chars: Optional[List[str]] = None,
    use_endlines: Optional[bool] = None,
    ignore_excluded: bool = True,
):
    r'''
    The `eds.sentences` matcher provides an alternative to spaCy's default
    `sentencizer`, aiming to overcome some of its limitations.

    Indeed, the `sentencizer` merely looks at period characters to detect the end of a
    sentence, a strategy that often fails in a clinical note settings. Our
    `eds.sentences` component also classifies end-of-lines as sentence boundaries if
    the subsequent token begins with an uppercase character, leading to slightly better
    performances.

    Moreover, the `eds.sentences` component use the output of the `eds.normalizer`
    and `eds.endlines` output by default when these components are added to the
    pipeline.

    Examples
    --------
    === "EDS-NLP"

        ```{ .python .no-check }
        import edsnlp, edsnlp.pipes as eds

        nlp = edsnlp.blank("eds")
        nlp.add_pipe(eds.sentences())  # same as nlp.add_pipe("eds.sentences")

        text = """Le patient est admis le 23 août 2021 pour une douleur à l'estomac
        Il lui était arrivé la même chose il y a deux ans."
        """

        doc = nlp(text)

        for sentence in doc.sents:
            print("<s>", sentence, "</s>")
        # Out: <s> Le patient est admis le 23 août 2021 pour une douleur à l'estomac
        # Out:  <\s>
        # Out: <s> Il lui était arrivé la même chose il y a deux ans. <\s>
        ```

    === "spaCy sentencizer"

        ```{ .python .no-check }
        import edsnlp, edsnlp.pipes as eds

        nlp = edsnlp.blank("eds")
        nlp.add_pipe("sentencizer")

        text = """Le patient est admis le 23 août 2021 pour une douleur à l'estomac"
        Il lui était arrivé la même chose il y a deux ans.
        """

        doc = nlp(text)

        for sentence in doc.sents:
            print("<s>", sentence, "</s>")
        # Out: <s> Le patient est admis le 23 août 2021 pour une douleur à l'estomac
        # Out: Il lui était arrivé la même chose il y a deux ans. <\s>
        ```

    Notice how EDS-NLP's implementation is more robust to ill-defined sentence endings.


    Parameters
    ----------
    nlp: PipelineProtocol
        The pipeline object.
    name: str
        The name of the component.
    punct_chars : Optional[List[str]]
        Punctuation characters.
    use_endlines : bool
        Whether to use endlines prediction.
    ignore_excluded : bool
        Whether to skip excluded tokens (requires the upstream `eds.normalizer` pipe).

    Authors and citation
    --------------------
    The `eds.sentences` component was developed by AP-HP's Data Science team.
    '''

    return SentenceSegmenter(
        nlp=nlp,
        name=name,
        punct_chars=punct_chars,
        use_endlines=use_endlines,
        ignore_excluded=ignore_excluded,
    )



from spacy.tokens import Token

if not Token.has_extension("excluded"):
    Token.set_extension("excluded", default=False)


def excluded_or_space_getter(t):
    return t.is_space or t.tag_ == "EXCLUDED"


if not Token.has_extension("excluded_or_space"):
    Token.set_extension(
        "excluded_or_space",
        getter=excluded_or_space_getter,
    )



import warnings
from typing import Any, Union

from confit import VisibleDeprecationWarning
from spacy.tokens import Doc, Span, Token


def deprecated_extension(name: str, new_name: str) -> None:
    msg = (
        f'The extension "{name}" is deprecated and will be '
        "removed in a future version. "
        f'Please use "{new_name}" instead.'
    )

    warnings.warn(msg, VisibleDeprecationWarning)


class deprecated_getter_factory:
    def __init__(self, name: str, new_name: str):
        self.name = name
        self.new_name = new_name

    def __call__(self, toklike: Union[Token, Span, Doc]) -> Any:
        n = f"{type(toklike).__name__}._.{self.name}"
        nn = f"{type(toklike).__name__}._.{self.new_name}"

        deprecated_extension(n, nn)

        return getattr(toklike._, self.new_name)



import pytest
import spacy
from spacy.lang.fr.lex_attrs import like_num


def test_eds_tokenizer_handles_long_text():
    text = """Témoin interne : + ; témoin externe : +
- Récepteurs aux œstrogènes : tous les élements sont marqués (3+).
- Récepteurs à la progestérone : 40% sont marqués (intensité 2+).
- Anti-Cerb B2: 0% des cellules carcinomateuses présentent
un marquage membranaire complet.
CONCLUSION`: ========
-Carcinome mammaire infiltrant du quadrant inféro-externe,
de 25 mm size de grade II de
malignité selon Elston et Ellis (3+2+1), sans composante
Score ACR5 de chaque coté`'.
On se donne rendez-vous pour le 23/11/1967.
"""
    nlp = spacy.blank("eds")
    tokens = nlp(text)
    assert (
        [t.text_with_ws for t in tokens]
        == """Témoin |interne |: |+ |; |témoin |externe |: |+|
|- |Récepteurs |aux |œstrogènes |: |tous |les |élements |sont |marqués |(|3|+|)|.|
|- |Récepteurs |à |la |progestérone |: |40|% |sont |marqués |(|intensité |2|+|)|.|
|- |Anti|-|Cerb |B|2|: |0|% |des |cellules |carcinomateuses |présentent|
|un |marquage |membranaire |complet|.|
|CONCLUSION|`|: |=|=|=|=|=|=|=|=|
|-|Carcinome |mammaire |infiltrant |du |quadrant |inféro|-|externe|,|
|de |25 |mm |size |de |grade |II |de|
|malignité |selon |Elston |et |Ellis |(|3|+|2|+|1|)|, |sans |composante|
|Score |ACR|5 |de |chaque |coté|`|'|.|
|On |se |donne |rendez|-|vous |pour |le |23|/|11|/|1967|.|
""".split(
            "|"
        )
    )


@pytest.mark.parametrize("word", ["onze", "onzième"])
def test_eds_lex_attrs_capitals(word):
    assert like_num(word)
    assert like_num(word.upper())


def test_eds_tokenizer_whitespace():
    nlp = spacy.blank("eds")
    tokenized = [(w.text, w.whitespace_) for w in nlp("Lorem\xA0Ipsum\tDolor Sit Amet")]
    assert tokenized == [
        ("Lorem", " "),
        ("Ipsum", ""),
        ("\t", ""),
        ("Dolor", " "),
        ("Sit", " "),
        ("Amet", ""),
    ]


def test_eds_tokenizer_numbers():
    nlp = spacy.blank("eds")
    tokenized = [(w.text, w.whitespace_) for w in nlp("Il fait 5.3/5.4mm")]
    assert tokenized == [
        ("Il", " "),
        ("fait", " "),
        ("5.3", ""),
        ("/", ""),
        ("5.4", ""),
        ("mm", ""),
    ]


def test_eds_tokenizer_exceptions():
    nlp = spacy.blank("eds")
    txt = "M. Gentil a un rhume, code ADICAP: B.H.HP.A7A0"
    tokenized = [(w.text, w.whitespace_) for w in nlp(txt)]
    assert tokenized == [
        ("M.", " "),
        ("Gentil", " "),
        ("a", " "),
        ("un", " "),
        ("rhume", ""),
        (",", " "),
        ("code", " "),
        ("ADICAP", ""),
        (":", " "),
        ("B.", ""),
        ("H.", ""),
        ("HP.", ""),
        ("A", ""),
        ("7", ""),
        ("A", ""),
        ("0", ""),
    ]



examples = [
    "1. Codification ADICAP : OHNP6071",
    "2. Codification ADICAP : CSGS7624",
    "3. Codification ADICAP : BSDETMA0",
    "4. Codification ADICAP : BSSG0503",
    "5. Codification ADICAP : BSGS0503",
    "6. Codification ADICAP : HSGS0G26",
    "7. Codification ADICAP : HSGS0150",
]


expected_outputs = [
    {
        "code": "OHNP6071",
        "sampling_mode": "PIECE OPERATOIRE AVEC EXERESE COMPLETE DE L'ORGANE",
        "technic": "HISTOLOGIE ET CYTOLOGIE PAR INCLUSION",
        "organ": "SYSTEME NERVEUX PERIPHERIQUE",
        "pathology": "PATHOLOGIE GÉNÉRALE NON TUMORALE",
        "pathology_type": (
            "THROMBOCYTOPENIE - HYPOPLASIE ACQUISE "
            "MEGACARYOCYTAIRE ET PLAQUETTAIRE (SAI)"
        ),
        "behaviour_type": "CARACTERES GENERAUX",
    },
    {
        "code": "CSGS7624",
        "sampling_mode": "CYTOPONCTION NON GUIDEE PAR IMAGERIE",
        "technic": "COLORATION SPECIALE - HISTO ET CYTOCHIMIE",
        "organ": "SEIN (ÉGALEMENT UTILISÉ CHEZ L'HOMME)",
        "pathology": "PATHOLOGIE GÉNÉRALE NON TUMORALE",
        "pathology_type": (
            "INFLAMMATION SUBAIGUE ET CHRONIQUE LEGERE TOXIQUE (ALCOOL, ETC)"
        ),
        "behaviour_type": (
            "MODIFICATION DE VOLUME, DYSTROPHIE, "
            "KYSTE, METAPLASIE ET DYSPLASIE ACQUISE"
        ),
    },
    {
        "code": "BSDETMA0",
        "sampling_mode": "BIOPSIE CHIRURGICALE",
        "technic": "COLORATION SPECIALE - HISTO ET CYTOCHIMIE",
        "organ": "ESTOMAC",
        "pathology": "PATHOLOGIE TUMORALE",
        "pathology_type": "THYMOME AVEC EMBOLES NEOPLASIQUES",
        "behaviour_type": "CANCER MÉTASTATIQUE",
    },
    {
        "code": "BSSG0503",
        "sampling_mode": "BIOPSIE CHIRURGICALE",
        "technic": "COLORATION SPECIALE - HISTO ET CYTOCHIMIE",
        "organ": "GANGLION LYMPHATIQUE",
        "pathology": "PATHOLOGIE PARTICULIERE DES ORGANES",
        "pathology_type": "HISTIOCYTOSE EN AMAS GANGLIONNAIRE",
        "behaviour_type": None,
    },
    {
        "code": "BSGS0503",
        "sampling_mode": "BIOPSIE CHIRURGICALE",
        "technic": "COLORATION SPECIALE - HISTO ET CYTOCHIMIE",
        "organ": "SEIN (ÉGALEMENT UTILISÉ CHEZ L'HOMME)",
        "pathology": None,
        "pathology_type": None,
        "behaviour_type": None,
    },
    {
        "code": "HSGS0G26",
        "sampling_mode": "HISTOPONCTION GUIDEE PAR IMAGERIE",
        "technic": "COLORATION SPECIALE - HISTO ET CYTOCHIMIE",
        "organ": "SEIN (ÉGALEMENT UTILISÉ CHEZ L'HOMME)",
        "pathology": "CYTOPATHOLOGIE",
        "pathology_type": (
            "CELLULES GLANDULAIRES ATYPIQUES "
            "EN FAVEUR D'UNE NATURE NEOPLASIQUE - SAI"
        ),
        "behaviour_type": "MATERIEL DE SIGNIFICATION INDETERMINEE",
    },
    {
        "code": "HSGS0150",
        "sampling_mode": "HISTOPONCTION GUIDEE PAR IMAGERIE",
        "technic": "COLORATION SPECIALE - HISTO ET CYTOCHIMIE",
        "organ": "SEIN (ÉGALEMENT UTILISÉ CHEZ L'HOMME)",
        "pathology": "PATHOLOGIE GÉNÉRALE NON TUMORALE",
        "pathology_type": "LESION D'INTERET PARTICULIER",
        "behaviour_type": "MALADIE INNEE ET GRANDE MALFORMATION EXTERNE",
    },
]


def test_scores(blank_nlp):

    blank_nlp.add_pipe("eds.adicap")

    for input, expected in zip(examples, expected_outputs):
        doc = blank_nlp(input)
        assert len(doc.ents) > 0
        assert doc.ents[0]._.adicap.dict() == expected



from typing import List, Optional

from spacy.tokens import Doc

from edsnlp.core import PipelineProtocol

from ...base import BaseComponent
from .fast_sentences import FastSentenceSegmenter
from .terms import punctuation


class SentenceSegmenter(BaseComponent):
    """
    Segments the Doc into sentences using a rule-based strategy,
    specific to AP-HP documents.

    Applies the same rule-based pipeline as spaCy's sentencizer,
    and adds a simple rule on the new lines : if a new line is followed by a
    capitalised word, then it is also an end of sentence.
    """

    def __init__(
        self,
        nlp: PipelineProtocol,
        name: Optional[str] = "sentences",
        punct_chars: Optional[List[str]] = None,
        use_endlines: bool = None,
        ignore_excluded: bool = True,
    ):
        """
        Parameters
        ----------
        nlp: PipelineProtocol
            The EDS-NLP pipeline
        name: Optional[str]
            The name of the component
        punct_chars: Optional[List[str]]
            Punctuation characters.
        use_endlines: bool
            Whether to use endlines prediction.
        ignore_excluded: bool
            Whether to ignore excluded tokens.
        """
        super().__init__(nlp, name)

        if punct_chars is None:
            punct_chars = punctuation

        self.fast_segmenter = FastSentenceSegmenter(
            vocab=nlp.vocab,
            punct_chars=punct_chars,
            use_endlines=use_endlines,
            ignore_excluded=ignore_excluded,
        )

    def __call__(self, doc: Doc):
        return self.fast_segmenter(doc)



import math
from pathlib import Path
from typing import List, Optional, Set, Tuple, Union

import tokenizers
import torch
from confit import validate_arguments
from foldedtensor import as_folded_tensor
from transformers import AutoModel, AutoTokenizer
from transformers import BitsAndBytesConfig as BitsAndBytesConfig_
from typing_extensions import Literal, TypedDict

from edsnlp import Pipeline
from edsnlp.pipes.trainable.embeddings.typing import WordEmbeddingComponent
from edsnlp.utils.span_getters import SpanGetterArg, get_spans

INITIAL_MAX_TOKENS_PER_DEVICE = 32 * 128
TransformerBatchInput = TypedDict(
    "TransformerBatchInput",
    {
        "input_ids": torch.LongTensor,
        "attention_mask": torch.BoolTensor,
        "token_window_indices": torch.LongTensor,
        "words_offsets": torch.LongTensor,
        "mask": torch.BoolTensor,
    },
)

BitsAndBytesConfig = validate_arguments(BitsAndBytesConfig_)


def compute_contextualization_scores(windows):
    ramp = torch.arange(0, windows.shape[1], 1)
    scores = (
        torch.min(ramp, windows.mask.sum(1, keepdim=True) - 1 - ramp)
        .clamp(min=0)
        .view(-1)
    )
    return scores


class Transformer(WordEmbeddingComponent[TransformerBatchInput]):
    """
    The `eds.transformer` component is a wrapper around HuggingFace's
    [transformers](https://huggingface.co/transformers/) library. If you are not
    familiar with transformers, a good way to start is the
    [Illustrated Transformer](https://jalammar.github.io/illustrated-transformer/)
    tutorial.

    Compared to using the raw Huggingface model, we offer a simple
    mechanism to split long documents into strided windows before feeding them to the
    model.

    Windowing
    ---------
    EDS-NLP's Transformer component splits long documents into smaller windows before
    feeding them to the model. This is done to avoid hitting the maximum number of
    tokens that can be processed by the model on a single device. The window size and
    stride can be configured using the `window` and `stride` parameters. The default
    values are 512 and 256 respectively, which means that the model will process windows
    of 512 tokens, each separated by 256 tokens. Whenever a token appears in multiple
    windows, the embedding of the "most contextualized" occurrence is used, i.e. the
    occurrence that is the closest to the center of its window.

    Here is an overview how this works to produce embeddings (shown in red) for each
    word of the document :

    <figure style="text-align: center" markdown>
    ![Transformer windowing](/assets/images/transformer-windowing.svg)
    </figure>

    Examples
    --------
    Here is an example of how to define a pipeline with a Transformer component:

    ```python
    import edsnlp, edsnlp.pipes as eds

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(
        eds.transformer(
            model="prajjwal1/bert-tiny",
            window=128,
            stride=96,
        ),
    )
    ```

    You can then compose this embedding with a task specific component such as
    `eds.ner_crf`.

    Parameters
    ----------
    nlp: PipelineProtocol
        The pipeline instance
    name: str
        The component name
    model: str
        The Huggingface model name or path
    window: int
        The window size to use when splitting long documents into smaller windows
        before feeding them to the Transformer model (default: 512 = 512 - 2)
    stride: int
        The stride (distance between windows) to use when splitting long documents into
        smaller windows: (default: 96)
    max_tokens_per_device: Union[int, Literal["auto"]]
        The maximum number of tokens that can be processed by the model on a single
        device. This does not affect the results but can be used to reduce the memory
        usage of the model, at the cost of a longer processing time.

        If "auto", the component will try to estimate the maximum number of tokens that
        can be processed by the model on the current device at a given time.
    span_getter: Optional[SpanGetterArg]
        Which spans of the document should be embedded. Defaults to the full document
        if None.
    """

    def __init__(
        self,
        nlp: Optional[Pipeline] = None,
        name: str = "transformer",
        *,
        model: Union[str, Path],
        window: int = 128,
        stride: int = 96,
        max_tokens_per_device: Union[int, Literal["auto"]] = "auto",
        span_getter: Optional[SpanGetterArg] = None,
        new_tokens: Optional[List[Tuple[str, str]]] = [],
        quantization: Optional[BitsAndBytesConfig] = None,
        **kwargs,
    ):
        super().__init__(nlp, name)
        self.transformer = AutoModel.from_pretrained(
            model,
            quantization_config=quantization,
            **kwargs,
        )
        self.tokenizer = AutoTokenizer.from_pretrained(model)
        self.window = window
        self.stride = stride
        self.output_size = self.transformer.config.hidden_size
        self.empty_word_embedding = torch.nn.Parameter(
            torch.randn(
                (1, self.output_size),
            )
        )
        self.cfg = {}
        self.max_tokens_per_device = max_tokens_per_device
        self._mem_per_unit = None
        self.span_getter = span_getter

        if new_tokens:
            self.tokenizer.add_tokens(sorted(set(t[1] for t in new_tokens)))
            original_normalizer = self.tokenizer.backend_tokenizer.normalizer
            self.tokenizer.backend_tokenizer.normalizer = (
                tokenizers.normalizers.Sequence(
                    [
                        *(
                            tokenizers.normalizers.Replace(
                                tokenizers.Regex(pattern), replacement
                            )
                            for pattern, replacement in new_tokens
                        ),
                        original_normalizer,
                    ]
                )
            )
            # and add a new entry to the model's embeddings
            self.transformer.resize_token_embeddings(
                max(self.tokenizer.vocab.values()) + 1
            )

    def to_disk(self, path, *, exclude: Optional[Set[str]]):
        repr_id = object.__repr__(self)
        if repr_id in exclude:
            return
        self.tokenizer.save_pretrained(path)
        self.transformer.save_pretrained(path)
        for param in self.transformer.parameters():
            exclude.add(object.__repr__(param))
        cfg = super().to_disk(path, exclude=exclude) or {}
        cfg["model"] = f"./{path.as_posix()}"
        return cfg

    def preprocess(self, doc):
        res = {
            "input_ids": [],
            "word_tokens": [],
            "word_lengths": [],
        }

        for span in get_spans(doc, self.span_getter):
            # Preprocess it using LayoutLMv3
            prep = self.tokenizer(
                span.text,
                is_split_into_words=False,
                return_offsets_mapping=True,
                return_attention_mask=False,
            )

            span_word_tokens, span_word_lengths = self.align_words_with_trf_tokens(
                span, prep.pop("offset_mapping")
            )
            res["input_ids"].append(prep["input_ids"])
            res["word_tokens"].append(span_word_tokens)
            res["word_lengths"].append(span_word_lengths)

        return res

    def collate(self, batch):
        # Flatten most of these arrays to process batches page per page and
        # not sample per sample

        S = self.stride
        W = self.window

        offset = 0
        window_max_size = 0
        window_count = 0
        windows = []
        windows_count_per_page = []
        for sample_input_ids in batch["input_ids"]:
            for span_input_ids in sample_input_ids:
                span_size = len(span_input_ids)
                num_span_windows = 1 + max(0, math.ceil((span_size - 2 - W) / S))
                windows.append(
                    [
                        [
                            offset + 0,
                            *range(
                                1 + offset + win_i * S,
                                1 + offset + min(win_i * S + W, span_size - 2),
                            ),
                            offset + len(span_input_ids) - 1,
                        ]
                        for win_i in range(0, num_span_windows)
                    ]
                )
                windows_count_per_page.append(len(windows[-1]))
                offset += len(span_input_ids)
                window_max_size = max(window_max_size, max(map(len, windows[-1])))
                window_count += len(windows[-1])

        windows = as_folded_tensor(
            windows,
            full_names=("sample", "window", "token"),
            data_dims=("window", "token"),
            dtype=torch.long,
        )
        indexer = torch.zeros(
            (0 if 0 in windows.shape else windows.max()) + 1, dtype=torch.long
        )

        # Sort each occurrence of an initial token by its contextualization score:
        # We can only use the amax reduction, so to retrieve the best occurrence, we
        # insert the index of the token output by the transformer inside the score
        # using a lexicographic approach
        # (score + index / n_tokens) ~ (score * n_tokens + index), taking the max,
        # and then retrieving the index of the token using the modulo operator.
        scores = compute_contextualization_scores(windows)
        scores = scores * len(scores) + torch.arange(len(scores))

        indexer.index_reduce_(
            dim=0,
            source=scores,
            index=windows.view(-1),
            reduce="amax",
        )
        indexer %= max(len(scores), 1)
        indexer = indexer.tolist()
        # Indexer: flattened, unpadded mapping tensor, that for each wordpiece in the
        # batch gives its position inside the flattened padded windows before/after the
        # transformer

        input_ids = as_folded_tensor(
            batch["input_ids"],
            data_dims=("subword",),
            full_names=("sample", "span", "subword"),
            dtype=torch.long,
        ).as_tensor()[windows]
        empty_id = input_ids.numel()

        word_indices = []
        word_offsets = []
        offset = 0
        indexer_offset = 0
        mask = []
        for sample_word_lengths, sample_word_tokens, sample_input_ids in zip(
            batch["word_lengths"], batch["word_tokens"], batch["input_ids"]
        ):
            for span_word_lengths, span_word_tokens, span_input_ids in zip(
                sample_word_lengths, sample_word_tokens, sample_input_ids
            ):
                offset = 0
                for length in span_word_lengths:
                    word_offsets.append(len(word_indices))
                    word_indices.extend(
                        [
                            indexer[indexer_offset + i] if i >= 0 else empty_id
                            for i in span_word_tokens[offset : offset + length]
                        ]
                    )
                    offset += length
                indexer_offset += len(span_input_ids)
                mask.append([True] * len(span_word_lengths))

        return {
            "input_ids": input_ids.as_tensor(),
            "input_mask": input_ids.mask,
            "word_indices": torch.as_tensor(word_indices),
            "word_offsets": torch.as_tensor(word_offsets),
            "mask": as_folded_tensor(
                mask,
                data_dims=(
                    "span",
                    "subword",
                ),
                full_names=("span", "subword"),
                dtype=torch.bool,
            ).as_tensor(),
        }

    def forward(self, batch):
        device = batch["input_ids"].device
        if len(batch["input_ids"]) == 0:
            return {
                "embeddings": torch.zeros(
                    (*batch["mask"].shape, self.output_size),
                    dtype=torch.float,
                    device=device,
                ),
                "mask": batch["mask"].clone(),
            }

        kwargs = dict(
            input_ids=batch["input_ids"],
            attention_mask=batch["input_mask"],
        )
        auto_batch_size = device.type == "cuda" and self.max_tokens_per_device == "auto"
        trial_idx = 1
        while True:
            total_tokens = batch["input_ids"].numel()
            if auto_batch_size:  # pragma: no cover
                max_tokens = INITIAL_MAX_TOKENS_PER_DEVICE
                torch.cuda.synchronize(device)
                total_mem = torch.cuda.get_device_properties(device).total_memory
                allocated_mem = torch.cuda.memory_allocated(device)
                torch.cuda.reset_peak_memory_stats(device)
                free_mem = total_mem - allocated_mem

                if self._mem_per_unit is not None:
                    max_tokens = max(1, int(free_mem // self._mem_per_unit))
            else:
                max_tokens = (
                    self.max_tokens_per_device
                    if self.max_tokens_per_device != "auto"
                    else INITIAL_MAX_TOKENS_PER_DEVICE
                )

            max_windows = max(1, max_tokens // batch["input_ids"].size(1))
            total_windows = batch["input_ids"].size(0)
            try:
                wordpiece_embeddings = [
                    self.transformer.base_model(
                        **{
                            k: None if v is None else v[offset : offset + max_windows]
                            for k, v in kwargs.items()
                        }
                    ).last_hidden_state
                    for offset in range(0, total_windows, max_windows)
                ]

                wordpiece_embeddings = (
                    torch.cat(wordpiece_embeddings, dim=0)
                    if len(wordpiece_embeddings) > 1
                    else wordpiece_embeddings[0]
                )

                if auto_batch_size:  # pragma: no cover
                    batch_mem = torch.cuda.max_memory_allocated(device)
                    current_mem_per_unit = batch_mem / min(total_tokens, max_tokens)
                    if self._mem_per_unit is None:
                        self._mem_per_unit = current_mem_per_unit
                    self._mem_per_unit = (
                        self._mem_per_unit * 0.7 + current_mem_per_unit * 0.3
                    )
                    torch.cuda.synchronize(device)  # Wait for all kernels to finish
            except RuntimeError as e:  # pragma: no cover
                if "out of memory" in str(e) and trial_idx <= 2:
                    print(
                        f"Out of memory: tried to fit {max_windows} "
                        f"in {free_mem / (1024 ** 3)} (try n°{trial_idx}/2)"
                    )
                    torch.cuda.empty_cache()
                    self._mem_per_unit = (free_mem / max_windows) * 1.5
                    trial_idx += 1
                    continue
                raise
            break

        mask = batch["mask"].clone()
        word_embeddings = torch.zeros(
            (mask.size(0), mask.size(1), wordpiece_embeddings.size(2)),
            dtype=torch.float,
            device=device,
        )
        embeddings_plus_empty = torch.cat(
            [
                wordpiece_embeddings.view(-1, wordpiece_embeddings.size(2)),
                self.empty_word_embedding,
            ],
            dim=0,
        )
        word_embeddings[mask] = torch.nn.functional.embedding_bag(
            input=batch["word_indices"],
            weight=embeddings_plus_empty,
            offsets=batch["word_offsets"],
        )
        return {
            "embeddings": word_embeddings,
            "mask": mask,
        }

    def align_words_with_trf_tokens(self, doc, trf_char_indices):
        token_i = 0
        n_trf_tokens = len(trf_char_indices)
        word_tokens = []
        word_lengths = [0] * (len(doc))
        doc_start = doc[0].idx if len(doc) else 0

        for word_i, word in enumerate(doc):
            length = 0
            word_begin = word.idx - doc_start
            word_end = word_begin + len(word)

            for j in range(token_i, n_trf_tokens):
                if trf_char_indices[j][1] <= word_begin:
                    token_i += 1
                elif trf_char_indices[j][0] >= word_end:
                    break
                else:
                    length += 1
                    word_tokens.append(j)
            if length > 0:
                word_lengths[word_i] = length
            else:
                word_tokens.append(-1)
                word_lengths[word_i] = 1

        return word_tokens, word_lengths



from typing import Optional

from spacy.tokens import Doc

from edsnlp.core import PipelineProtocol

from ...base import BaseComponent
from .accents.accents import AccentsConverter
from .pollution.pollution import PollutionTagger
from .quotes.quotes import QuotesConverter
from .remove_lowercase.factory import remove_lowercase
from .spaces.spaces import SpacesTagger


class Normalizer(BaseComponent):
    """
    Normalisation pipeline. Modifies the `NORM` attribute,
    acting on five dimensions :

    - `lowercase`: using the default `NORM`
    - `accents`: deterministic and fixed-length normalisation of accents.
    - `quotes`: deterministic and fixed-length normalisation of quotation marks.
    - `spaces`: "removal" of spaces tokens (via the tag_ attribute).
    - `pollution`: "removal" of pollutions (via the tag_ attribute).

    Parameters
    ----------
    nlp : Optional[PipelineProtocol]
        The pipeline object.
    name : Optional[str]
        The name of the component.
    lowercase : bool
        Whether to remove case.
    accents : Optional[Accents]
        Optional `Accents` object.
    quotes : Optional[Quotes]
        Optional `Quotes` object.
    spaces : Optional[Spaces]
        Optional `Spaces` object.
    pollution : Optional[Pollution]
        Optional `Pollution` object.
    """

    def __init__(
        self,
        nlp: Optional[PipelineProtocol],
        name: Optional[str] = "normalizer",
        *,
        lowercase: bool = False,
        accents: Optional[AccentsConverter] = None,
        quotes: Optional[QuotesConverter] = None,
        spaces: Optional[SpacesTagger] = None,
        pollution: Optional[PollutionTagger] = None,
    ):
        super().__init__(nlp, name)

        self.lowercase = lowercase
        self.accents = accents
        self.quotes = quotes
        self.spaces = spaces
        self.pollution = pollution

    def __call__(self, doc: Doc) -> Doc:
        """
        Apply the normalisation pipeline, one component at a time.

        Parameters
        ----------
        doc : Doc
            spaCy `Doc` object

        Returns
        -------
        Doc
            Doc object with `NORM` attribute modified
        """
        if not self.lowercase:
            remove_lowercase(doc)
        if self.accents is not None:
            self.accents(doc)
        if self.quotes is not None:
            self.quotes(doc)
        if self.spaces is not None:
            self.spaces(doc)
        if self.pollution is not None:
            self.pollution(doc)
        return doc



from edsnlp.core import registry

from .matcher import GenericMatcher

DEFAULT_CONFIG = dict(
    terms=None,
    regex=None,
    attr="TEXT",
    ignore_excluded=False,
    ignore_space_tokens=False,
    term_matcher="exact",
    term_matcher_config={},
    span_setter={"ents": True},
)

create_component = registry.factory.register(
    "eds.matcher",
    assigns=["doc.ents", "doc.spans"],
    deprecated=["matcher"],
)(GenericMatcher)



import logging
from datetime import datetime

import pandas as pd
import pytest
import spacy
from helpers import make_nlp
from pytest import fixture

import edsnlp

logging.basicConfig(level=logging.INFO)


@fixture(scope="session", params=["eds", "fr"])
def lang(request):
    return request.param


@fixture(scope="session")
def nlp(lang):
    return make_nlp(lang)


@fixture(scope="session")
def nlp_eds():
    return make_nlp("eds")


@fixture
def blank_nlp(lang):
    if lang == "eds":
        model = spacy.blank("eds")
    else:
        model = edsnlp.blank("fr")
    model.add_pipe("eds.sentences")
    return model


def make_ml_pipeline():
    nlp = edsnlp.blank("eds")
    nlp.add_pipe("eds.sentences", name="sentences")
    nlp.add_pipe(
        "eds.transformer",
        name="transformer",
        config=dict(
            model="prajjwal1/bert-tiny",
            window=128,
            stride=96,
        ),
    )
    nlp.add_pipe(
        "eds.ner_crf",
        name="ner",
        config=dict(
            embedding=nlp.get_pipe("transformer"),
            mode="independent",
            target_span_getter=["ents", "ner-preds"],
            span_setter="ents",
        ),
    )
    ner = nlp.get_pipe("ner")
    ner.update_labels(["PERSON", "GIFT"])
    return nlp


@fixture()
def ml_nlp():
    return make_ml_pipeline()


@fixture(scope="session")
def frozen_ml_nlp():
    return make_ml_pipeline()


@fixture()
def text():
    return (
        "Le patient est admis pour des douleurs dans le bras droit, "
        "mais n'a pas de problème de locomotion. "
        "Historique d'AVC dans la famille. pourrait être un cas de rhume.\n"
        "NBNbWbWbNbWbNBNbNbWbWbNBNbWbNbNbWbNBNbWbNbNBWbWbNbNbNBWbNbWbNbWBNb"
        "NbWbNbNBNbWbWbNbWBNbNbWbNBNbWbWbNb\n"
        "Pourrait être un cas de rhume.\n"
        "Motif :\n"
        "Douleurs dans le bras droit.\n"
        "ANTÉCÉDENTS\n"
        "Le patient est déjà venu pendant les vacances\n"
        "d'été.\n"
        "Pas d'anomalie détectée."
    )


@fixture
def doc(nlp, text):
    return nlp(text)


@fixture
def blank_doc(blank_nlp, text):
    return blank_nlp(text)


@fixture
def df_notes():
    N_LINES = 100
    notes = pd.DataFrame(
        data={
            "note_id": list(range(N_LINES)),
            "note_text": N_LINES * [text],
            "note_datetime": N_LINES * [datetime.today()],
        }
    )

    return notes


def make_df_note(text, module):
    from pyspark.sql import types as T
    from pyspark.sql.session import SparkSession

    data = [(i, i // 5, text, datetime(2021, 1, 1)) for i in range(20)]

    if module == "pandas":
        return pd.DataFrame(
            data=data, columns=["note_id", "person_id", "note_text", "note_datetime"]
        )

    note_schema = T.StructType(
        [
            T.StructField("note_id", T.IntegerType()),
            T.StructField("person_id", T.IntegerType()),
            T.StructField("note_text", T.StringType()),
            T.StructField(
                "note_datetime",
                T.TimestampType(),
            ),
        ]
    )

    spark = SparkSession.builder.getOrCreate()
    notes = spark.createDataFrame(data=data, schema=note_schema)
    if module == "pyspark":
        return notes

    if module == "koalas":
        try:
            import databricks.koalas  # noqa F401
        except ImportError:
            pytest.skip("Koalas not installed")
        return notes.to_koalas()


@fixture
def df_notes_pandas(text):
    return make_df_note(text, "pandas")


@fixture
def df_notes_pyspark(text):
    return make_df_note(text, "pyspark")


@fixture
def df_notes_koalas(text):
    return make_df_note(text, "koalas")


@fixture
def run_in_test_dir(request, monkeypatch):
    monkeypatch.chdir(request.fspath.dirname)



"""
Process ADICAP codes
Thésaurus de la codification ADICAP - Index raisonné des lésions
source : https://smt.esante.gouv.fr/terminologie-adicap/

"""

import gzip
import json
import re
from pathlib import Path

import pandas as pd
import typer


def parse_each_dict(df, dictionaryCode: str):
    d_spec = df.query(f"dictionaryCode=='{dictionaryCode}'")
    d_spec.fillna("", inplace=True)

    decode_d_spec = {}

    for code, label, anatomyCode in d_spec[["code", "label", "anatomyCode"]].values:
        if dictionaryCode == "D5":
            if re.match(r"[0-9]{4}", code) is None:
                decode_d_spec[str(anatomyCode) + str(code)] = label
        else:
            decode_d_spec[str(anatomyCode) + str(code)] = label

    d_value = decode_d_spec.pop(dictionaryCode)

    return dict(label=d_value, codes=decode_d_spec)


def get_decode_dict(df, dict_keys=["D1", "D2", "D3", "D4", "D5", "D6", "D7"]):
    decode_dict = {}
    for key in dict_keys:

        decode_dict[key] = parse_each_dict(df, dictionaryCode=key)

    return decode_dict


def run(
    raw: Path = typer.Argument(..., help="Path to the raw file"),
    output: Path = typer.Option(
        "edsnlp/resources/adicap.json.gz", help="Path to the output CSV table."
    ),
) -> None:
    """
    Convenience script to automatically process the ADICAP codes
    into a processable file.
    """

    df = pd.read_excel(
        raw,
        sheet_name="rawdatas",
        header=0,
    )

    decode_dict = get_decode_dict(df)

    typer.echo(f"Saving to {output}")

    with gzip.open(output, "w") as f:
        f.write(json.dumps(decode_dict).encode("utf-8"))

    typer.echo("Done !")


if __name__ == "__main__":
    typer.run(run)



import bisect
from typing import List, Optional, Union

import pandas as pd
from rich.console import Console
from rich.table import Table
from spacy.tokens import Doc, Token

from edsnlp.core import PipelineProtocol
from edsnlp.matchers.utils import get_text
from edsnlp.pipes.qualifiers.base import get_qualifier_extensions
from edsnlp.utils.extensions import rgetattr


class QuickExample:
    def __init__(self, nlp: PipelineProtocol, extensions: List[str] = []):
        self.nlp = nlp
        self.qualifiers = get_qualifier_extensions(nlp)
        self.extensions = extensions

    def __call__(
        self, object: Union[str, Doc], as_dataframe: bool = False
    ) -> Optional[pd.DataFrame]:
        """
        Displays the text and a table of entities

        Parameters
        ----------
        as_dataframe : bool, optional
            If true, returns the table as a DataFrame instead of displaying it,
            by default False

        Returns
        -------
        Optional[pd.DataFrame]
            The DataFrame describing the document
        """
        if isinstance(object, str):
            self.txt = object
            self.doc = self.nlp(object)
        elif isinstance(object, Doc):
            self.txt = object.text
            self.doc = object
        self.get_ents()
        self.get_ents_interval()
        self.get_text()
        return self.display(as_dataframe=as_dataframe)

    def get_ents(self):
        all_spans = {k: list(s) for k, s in self.doc.spans.items() if s}
        all_spans["ents"] = list(self.doc.ents)

        ents = []

        for key, spans in all_spans.items():
            for span in list(spans):
                if span in all_spans["ents"]:
                    all_spans["ents"].remove(span)
                start, end = span.start, span.end
                text = get_text(span, attr="TEXT", ignore_excluded=False)
                ent = dict(
                    key=key,
                    start=start,
                    end=end,
                    text=text,
                )
                for name, extension in self.qualifiers.items():
                    ent[name] = rgetattr(span, extension)
                for extension in self.extensions:
                    ent[extension] = rgetattr(span, extension)
                ents.append(ent)

        self.ents = ents

    def get_ents_interval(self):
        """
        From the list of all entities, removes overlapping spans
        """

        intervals = []
        for ent in self.ents:
            interval = (ent["start"], ent["end"])
            istart, iend = interval

            i = bisect.bisect_right(intervals, (iend, len(self.doc) + 1))

            for idx, (start, end) in enumerate(intervals[:i]):
                if end > istart:
                    interval = (start, iend)
                    del intervals[idx]
                    break

            bisect.insort(intervals, interval)

        self.intervals = intervals

    def is_ent(self, tok: Token) -> bool:
        """
        Check if the provided Token is part of an entity

        Parameters
        ----------
        tok : Token
            A spaCy Token

        Returns
        -------
        bool
            True if `tok` is part of an entity
        """
        for interval in self.intervals:
            if (tok.i >= interval[0]) and (tok.i < interval[1]):
                return True
        return False

    def get_text(self) -> None:
        """
        Adds bold tags to `self.text`
        """
        text = []
        for tok in self.doc:
            raw_tok_text = tok.text + tok.whitespace_
            tok_text = (
                f"[bold]{raw_tok_text}[not bold]" if self.is_ent(tok) else raw_tok_text
            )
            text.append(tok_text)
        self.text = "".join(text)

    def display(self, as_dataframe: bool = False) -> Optional[pd.DataFrame]:
        """
        Displays the text and a table of entities

        Parameters
        ----------
        as_dataframe : bool, optional
            If true, returns the table as a DataFrame instead of displaying it,
            by default False

        Returns
        -------
        Optional[pd.DataFrame]
            The DataFrame describing the document
        """

        console = Console()

        table = Table(title=self.text + "\n")

        headers = ["Entity", "Source"] + list(self.qualifiers.keys()) + self.extensions

        if not as_dataframe:
            [table.add_column(h) for h in headers]

            for ent in self.ents:
                table.add_row(
                    ent["text"],
                    ent["key"],
                    *(
                        "[green]" + str(ent[q]) if ent[q] else "[red]" + str(ent[q])
                        for q in self.qualifiers
                    ),
                    *(str(ent[extension]) for extension in self.extensions),
                )

            console.print(table)

        else:
            df = pd.DataFrame(
                [
                    [
                        ent["text"],
                        ent["key"],
                        *(ent[q] for q in list(self.qualifiers.keys())),
                        *(ent[e] for e in self.extensions),
                    ]
                    for ent in self.ents
                ],
                columns=headers,
            )

            console.print(self.text)
            return df



import copy
import itertools
from collections import defaultdict
from typing import (
    Any,
    Dict,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    TypeVar,
    Union,
)

T = TypeVar("T")


def ld_to_dl(ld: Iterable[Mapping[str, T]]) -> Dict[str, List[T]]:
    """
    Convert a list of dictionaries to a dictionary of lists

    Parameters
    ----------
    ld: Iterable[Mapping[str, T]]
        The list of dictionaries

    Returns
    -------
    Dict[str, List[T]]
        The dictionary of lists
    """
    ld = list(ld)
    return {k: [dic.get(k) for dic in ld] for k in (ld[0] if len(ld) else ())}


def dl_to_ld(dl: Mapping[str, Sequence[Any]]) -> Iterator[Dict[str, Any]]:
    """
    Convert a dictionary of lists to a list of dictionaries

    Parameters
    ----------
    dl: Mapping[str, Sequence[Any]]
        The dictionary of lists

    Returns
    -------
    List[Dict[str, Any]]
        The list of dictionaries
    """
    return (dict(zip(dl, t)) for t in zip(*dl.values()))


FLATTEN_TEMPLATE = """\
def flatten(root):
    res={}
    return res
"""


def _discover_scheme(obj):
    keys = defaultdict(lambda: [])

    def rec(current, path):
        if not isinstance(current, dict):
            keys[id(current)].append(path)
            return
        for key, value in current.items():
            if not key.startswith("$"):
                rec(value, (*path, key))

    rec(obj, ())

    code = FLATTEN_TEMPLATE.format(
        "{"
        + "\n".join(
            "{}: root{},".format(
                repr("|".join(map("/".join, key_list))),
                "".join(f"[{repr(k)}]" for k in key_list[0]),
            )
            for key_list in keys.values()
        )
        + "}"
    )
    return code


class batch_compress_dict:
    """
    Compress a sequence of dictionaries in which values that occur multiple times are
    deduplicated. The corresponding keys will be merged into a single string using
    the "|" character as a separator.
    This is useful to preserve referential identities when decompressing the dictionary
    after it has been serialized and deserialized.

    Parameters
    ----------
    seq: Iterable[Dict[str, Any]]
        Sequence of dictionaries to compress
    """

    __slots__ = ("flatten", "seq")

    def __init__(self, seq: Optional[Iterable[Dict[str, Any]]] = None):
        self.seq = seq
        self.flatten = None

    def __iter__(self):
        return batch_compress_dict(iter(self.seq))

    # def __getstate__(self):
    #     return {"seq": self.seq}

    # def __setstate__(self, state):
    #     self.seq = state["seq"]
    #     self.flatten = None

    def __call__(self, item):
        exec_result = {}
        if self.flatten is None:
            exec(_discover_scheme(item), {}, exec_result)
            self.flatten = exec_result["flatten"]
        return self.flatten(item)

    def __next__(self) -> Dict[str, List]:
        return self(next(self.seq))


def decompress_dict(seq: Union[Iterable[Dict[str, Any]], Dict[str, Any]]):
    """
    Decompress a dictionary of lists into a sequence of dictionaries.
    This function assumes that the dictionary structure was obtained using the
    `batch_compress_dict` class.
    Keys that were merged into a single string using the "|" character as a separator
    will be split into a nested dictionary structure.

    Parameters
    ----------
    seq: Union[Iterable[Dict[str, Any]], Dict[str, Any]]
        The dictionary to decompress or a sequence of dictionaries to decompress

    Returns
    -------

    """
    obj = ld_to_dl(seq) if isinstance(seq, Sequence) else seq
    res = {}
    for key, value in obj.items():
        for path in key.split("|"):
            current = res
            parts = path.split("/")
            for part in parts[:-1]:
                current = current.setdefault(part, {})
            current[parts[-1]] = value
    return res


def batchify(
    iterable: Iterable[T],
    batch_size: int,
    drop_last: bool = False,
) -> Iterable[List[T]]:
    """
    Yields batch that contain at most `batch_size` elements.
    If an item contains more than `batch_size` elements, it will be yielded as a single
    batch.

    Parameters
    ----------
    iterable
    batch_size
    drop_last
    """
    batch = []
    for item in iterable:
        if len(batch) >= batch_size:
            yield batch
            batch = []
        batch.append(item)
    if len(batch) > 0 and not drop_last:
        yield batch


def get_attr_item(base, attr):
    try:
        return base[attr]
    except (KeyError, TypeError):
        return getattr(base, attr)


def split_names(names):
    _names = []
    for part in names.split("."):
        try:
            _names.append(int(part))
        except ValueError:
            _names.append(part)
    return _names


def get_deep_attr(base, names):
    if isinstance(names, str):
        names = split_names(names)
    if len(names) == 0:
        return base
    [current, *remaining] = names
    return get_deep_attr(get_attr_item(base, current), remaining)


def set_attr_item(base, attr, val):
    try:
        base[attr] = val
    except (KeyError, TypeError):
        setattr(base, attr, val)
    return base


def set_deep_attr(base, names, val):
    if isinstance(names, str):
        names = split_names(names)
    if len(names) == 0:
        return val
    if len(names) == 1:
        if isinstance(base, (dict, list)):
            base[names[0]] = val
        else:
            setattr(base, names[0], val)
    [current, *remaining] = names
    attr = base[current] if isinstance(base, (dict, list)) else getattr(base, current)
    try:
        set_deep_attr(attr, remaining, val)
    except TypeError:
        new_attr = list(attr)
        set_deep_attr(new_attr, remaining, val)
        return set_attr_item(base, current, tuple(new_attr))
    return base


class multi_tee:
    """
    Makes copies of an iterable such that every iteration over it
    starts from 0. If the iterable is a sequence (list, tuple), just returns
    it since every iter() over the object restart from the beginning
    """

    def __new__(cls, iterable):
        if isinstance(iterable, Sequence):
            return iterable
        return super().__new__(cls)

    def __init__(self, iterable):
        self.main, self.copy = itertools.tee(iterable)

    def __iter__(self):
        if self.copy is not None:
            it = self.copy
            self.copy = None
            return it
        return copy.copy(self.main)


class FrozenDict(dict):
    """
    Copied from `spacy.util.SimpleFrozenDict` to ensure compatibility.


    """

    def __init__(self, *args, error: str = None, **kwargs) -> None:
        """Initialize the frozen dict. Can be initialized with pre-defined
        values.

        error (str): The error message when user tries to assign to dict.
        """
        from spacy import Errors

        if error is None:
            error = Errors.E095
        super().__init__(*args, **kwargs)
        self.error = error

    def __setitem__(self, key, value):
        raise NotImplementedError(self.error)

    def pop(self, key, default=None):
        raise NotImplementedError(self.error)

    def update(self, other):
        raise NotImplementedError(self.error)


class FrozenList(list):
    """
    Copied from `spacy.util.SimpleFrozenDict` to ensure compatibility
    """

    def __init__(self, *args, error: str = None) -> None:
        """Initialize the frozen list.

        error (str): The error message when user tries to mutate the list.
        """
        from spacy import Errors

        if error is None:
            error = Errors.E927
        self.error = error
        super().__init__(*args)

    def append(self, *args, **kwargs):
        raise NotImplementedError(self.error)

    def clear(self, *args, **kwargs):
        raise NotImplementedError(self.error)

    def extend(self, *args, **kwargs):
        raise NotImplementedError(self.error)

    def insert(self, *args, **kwargs):
        raise NotImplementedError(self.error)

    def pop(self, *args, **kwargs):
        raise NotImplementedError(self.error)

    def remove(self, *args, **kwargs):
        raise NotImplementedError(self.error)

    def reverse(self, *args, **kwargs):
        raise NotImplementedError(self.error)

    def sort(self, *args, **kwargs):
        raise NotImplementedError(self.error)


class FrozenNamespace:
    def __init__(self, items) -> None:
        self.__dict__.update(items)

    def __setattr__(self, key, value):
        raise NotImplementedError("You cannot assign to a frozen namespace")

    def __delattr__(self, key):
        raise NotImplementedError(
            "You cannot delete an attribute from a frozen namespace"
        )


def flatten_once(items):
    for item in items:
        if isinstance(item, list):
            yield from item
        else:
            yield item


def flatten(items):
    for item in items:
        if isinstance(item, list):
            yield from flatten(item)
        else:
            yield item



import os
from pathlib import Path

import mkdocs.config
import mkdocs.plugins
import mkdocs.structure
import mkdocs.structure.files
import mkdocs.structure.nav
import mkdocs.structure.pages
import regex
from mkdocs.config.defaults import MkDocsConfig

from docs.scripts.autorefs.plugin import AutorefsPlugin

try:
    from importlib.metadata import entry_points
except ImportError:
    from importlib_metadata import entry_points


def exclude_file(name):
    return name.startswith("assets/fragments/")


# Add the files from the project root

VIRTUAL_FILES = {}
REFERENCE_TEMPLATE = """
# `{ident}`
::: {ident}
    options:
        show_source: false
"""


@mkdocs.plugins.event_priority(1000)
def on_config(config: MkDocsConfig):
    for event_name, events in config.plugins.events.items():
        for event in list(events):
            if "autorefs" in str(event):
                print("REMOVING EVENT", event_name, event)
                events.remove(event)
    old_plugin = config["plugins"]["autorefs"]
    plugin_config = dict(old_plugin.config)
    plugin = AutorefsPlugin()
    config.plugins["autorefs"] = plugin
    config["plugins"]["autorefs"] = plugin
    plugin.load_config(plugin_config)


def on_files(files: mkdocs.structure.files.Files, config: mkdocs.config.Config):
    """
    Recursively the navigation of the mkdocs config
    and recursively content of directories of page that point
    to directories.

    Parameters
    ----------
    config: mkdocs.config.Config
        The configuration object
    kwargs: dict
        Additional arguments
    """

    root = Path("edsnlp")
    reference_nav = []
    for path in sorted(root.rglob("*.py")):
        module_path = path.relative_to(root.parent).with_suffix("")
        doc_path = Path("reference") / path.relative_to(root.parent).with_suffix(".md")
        # full_doc_path = Path("docs/reference/") / doc_path
        parts = list(module_path.parts)
        current = reference_nav
        for part in parts[:-1]:
            sub = next((item[part] for item in current if part in item), None)
            if sub is None:
                current.append({part: []})
                sub = current[-1][part]
            current = sub
        if parts[-1] == "__init__":
            parts = parts[:-1]
            doc_path = doc_path.with_name("index.md")
            current.append({"index.md": str(doc_path)})
        elif parts[-1] == "__main__":
            continue
        else:
            current.append({parts[-1]: str(doc_path)})
        ident = ".".join(parts)
        os.makedirs(doc_path.parent, exist_ok=True)
        VIRTUAL_FILES[str(doc_path)] = REFERENCE_TEMPLATE.format(ident=ident)

    for item in config["nav"]:
        if not isinstance(item, dict):
            continue
        key = next(iter(item.keys()))
        if not isinstance(item[key], str):
            continue
        if item[key].strip("/") == "reference":
            item[key] = reference_nav

    VIRTUAL_FILES["contributing.md"] = Path("contributing.md").read_text()
    VIRTUAL_FILES["changelog.md"] = Path("changelog.md").read_text()

    return mkdocs.structure.files.Files(
        [file for file in files if not exclude_file(file.src_path)]
        + [
            mkdocs.structure.files.File(
                file,
                config["docs_dir"],
                config["site_dir"],
                config["use_directory_urls"],
            )
            for file in VIRTUAL_FILES
        ]
    )


def on_nav(nav, config, files):
    def rec(node):
        if isinstance(node, list):
            return [rec(item) for item in node]
        if node.is_section and node.title == "Code Reference":
            return
        if isinstance(node, mkdocs.structure.nav.Navigation):
            return rec(node.items)
        if isinstance(node, mkdocs.structure.nav.Section):
            if (
                len(node.children)
                and node.children[0].is_page
                and node.children[0].is_index
            ):
                first = node.children[0]
                link = mkdocs.structure.nav.Link(
                    title=first.title,
                    url=first.url,
                )
                link.is_index = True
                first.title = "Overview"
                node.children.insert(0, link)
            return rec(node.children)

    rec(nav.items)


def on_page_read_source(page, config):
    if page.file.src_path in VIRTUAL_FILES:
        return VIRTUAL_FILES[page.file.src_path]
    return None


HREF_REGEX = (
    r"(?<=<\s*(?:a[^>]*href|img[^>]*src)=)"
    r'(?:"([^"]*)"|\'([^\']*)|[ ]*([^ =>]*)(?![a-z]+=))'
)
# Maybe find something less specific ?
PIPE_REGEX = r"(?<![a-zA-Z0-9._-])eds[.]([a-zA-Z0-9._-]*)(?![a-zA-Z0-9._-])"

HTML_PIPE_REGEX = r"""(?x)
(?<![a-zA-Z0-9._-])
<span[^>]*>eds<\/span>
<span[^>]*>[.]<\/span>
<span[^>]*>([a-zA-Z0-9._-]*)<\/span>
(?![a-zA-Z0-9._-])
"""


@mkdocs.plugins.event_priority(-1000)
def on_post_page(
    output: str,
    page: mkdocs.structure.pages.Page,
    config: mkdocs.config.Config,
):
    """
    1. Replace absolute paths with path relative to the rendered page
       This must be performed after all other plugins have run.
    2. Replace component names with links to the component reference

    Parameters
    ----------
    output
    page
    config

    Returns
    -------

    """

    autorefs: AutorefsPlugin = config["plugins"]["autorefs"]
    spacy_factories_entry_points = {
        ep.name: ep.value
        for ep in (
            *entry_points()["spacy_factories"],
            *entry_points()["edsnlp_factories"],
        )
    }

    def replace_component(match):
        full_group = match.group(0)
        name = "eds." + match.group(1)
        ep = spacy_factories_entry_points.get(name)
        preceding = output[match.start(0) - 50 : match.start(0)]
        if ep is not None and "DEFAULT:" not in preceding:
            try:
                url = autorefs.get_item_url(ep.replace(":", "."))
            except KeyError:
                pass
            else:
                return f"<a href={url}>{name}</a>"
        return full_group

    def replace_link(match):
        relative_url = url = match.group(1) or match.group(2) or match.group(3)
        page_url = os.path.join("/", page.file.url)
        if url.startswith("/"):
            relative_url = os.path.relpath(url, page_url)
        return f'"{relative_url}"'

    # Replace absolute paths with path relative to the rendered page
    output = regex.sub(PIPE_REGEX, replace_component, output)
    output = regex.sub(HTML_PIPE_REGEX, replace_component, output)
    output = regex.sub(HREF_REGEX, replace_link, output)

    return output



"""`eds.hemiplegia` pipeline"""
from typing import Any, Dict, List, Optional, Union

from edsnlp.core import PipelineProtocol
from edsnlp.pipes.base import SpanSetterArg

from ..base import DisorderMatcher
from .patterns import default_patterns


class HemiplegiaMatcher(DisorderMatcher):
    """
    The `eds.hemiplegia` pipeline component extracts mentions of hemiplegia.

    ??? info "Details of the used patterns"
        ```{ .python .no-check }
        # fmt: off
        --8<-- "edsnlp/pipes/ner/disorders/hemiplegia/patterns.py"
        # fmt: on
        ```

    Extensions
    ----------
    On each span `span` that match, the following attributes are available:

    - `span._.detailed_status`: set to `"PRESENT"`

    Examples
    --------
    ```python
    import edsnlp, edsnlp.pipes as eds

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(eds.sentences())
    nlp.add_pipe(
        eds.normalizer(
            accents=True,
            lowercase=True,
            quotes=True,
            spaces=True,
            pollution=dict(
                information=True,
                bars=True,
                biology=True,
                doctors=True,
                web=True,
                coding=True,
                footer=True,
            ),
        ),
    )
    nlp.add_pipe(eds.hemiplegia())
    ```

    Below are a few examples:

    --8<-- "docs/assets/fragments/hemiplegia-examples.md"

    Parameters
    ----------
    nlp : Optional[PipelineProtocol]
        The pipeline
    name : Optional[str]
        The name of the component
    patterns: Union[Dict[str, Any], List[Dict[str, Any]]]
        The patterns to use for matching
    label : str
        The label to use for the `Span` object and the extension
    span_setter : SpanSetterArg
        How to set matches on the doc

    # Authors and citation

    The `eds.hemiplegia` component was developed by AP-HP's Data Science team with a
    team of medical experts. A paper describing in details the development of those
    components is being drafted and will soon be available.
    """

    def __init__(
        self,
        nlp: Optional[PipelineProtocol],
        name: str = "hemiplegia",
        *,
        patterns: Union[Dict[str, Any], List[Dict[str, Any]]] = default_patterns,
        label: str = "hemiplegia",
        span_setter: SpanSetterArg = {"ents": True, "hemiplegia": True},
    ):
        super().__init__(
            nlp=nlp,
            name=name,
            patterns=patterns,
            label=label,
            span_setter=span_setter,
        )



from typing import Any, Dict, List, Optional

from spacy.tokens import Doc, Span
from typing_extensions import Literal

from edsnlp.core import PipelineProtocol
from edsnlp.matchers.phrase import EDSPhraseMatcher
from edsnlp.matchers.regex import RegexMatcher
from edsnlp.matchers.simstring import SimstringMatcher
from edsnlp.matchers.utils import Patterns
from edsnlp.pipes.base import BaseNERComponent, SpanSetterArg


class GenericMatcher(BaseNERComponent):
    r"""
    EDS-NLP simplifies the matching process by exposing a `eds.matcher` component
    that can match on terms or regular expressions.

    Examples
    --------
    Let us redefine the pipeline :

    ```python
    import edsnlp, edsnlp.pipes as eds

    nlp = edsnlp.blank("eds")

    terms = dict(
        covid=["coronavirus", "covid19"],  # (1)
        patient="patient",  # (2)
    )

    regex = dict(
        covid=r"coronavirus|covid[-\s]?19|sars[-\s]cov[-\s]2",  # (3)
    )

    nlp.add_pipe(
        eds.matcher(
            terms=terms,
            regex=regex,
            attr="LOWER",
            term_matcher="exact",
            term_matcher_config={},
        ),
    )
    ```

    1. Every key in the `terms` dictionary is mapped to a concept.
    2. The `eds.matcher` pipeline expects a list of expressions, or a single expression.
    3. We can also define regular expression patterns.

    This snippet is complete, and should run as is.

    Patterns, be they `terms` or `regex`, are defined as dictionaries where keys become
     the label of the extracted entities. Dictionary values are either a single
     expression or a list of expressions that match the concept.

    Parameters
    ----------
    nlp : PipelineProtocol
        The pipeline object.
    name: str
        The name of the component.
    terms : Optional[Patterns]
        A dictionary of terms.
    regex : Optional[Patterns]
        A dictionary of regular expressions.
    attr : str
        The default attribute to use for matching.
        Can be overridden using the `terms` and `regex` configurations.
    ignore_excluded : bool
        Whether to skip excluded tokens (requires an upstream
        pipeline to mark excluded tokens).
    ignore_space_tokens: bool
        Whether to skip space tokens during matching.

        You won't be able to match on newlines if this is enabled and
        the "spaces"/"newline" option of `eds.normalizer` is enabled (by default).
    term_matcher : Literal["exact", "simstring"]
        The matcher to use for matching phrases ?
        One of (exact, simstring)
    term_matcher_config : Dict[str,Any]
        Parameters of the matcher class
    span_setter : SpanSetterArg
        How to set the spans in the doc.

    Authors and citation
    --------------------
    The `eds.matcher` pipeline was developed by AP-HP's Data Science team.
    """

    def __init__(
        self,
        nlp: PipelineProtocol,
        name: Optional[str] = "matcher",
        *,
        terms: Optional[Patterns] = None,
        regex: Optional[Patterns] = None,
        attr: str = "TEXT",
        ignore_excluded: bool = False,
        ignore_space_tokens: bool = False,
        term_matcher: Literal["exact", "simstring"] = "exact",
        term_matcher_config: Dict[str, Any] = {},
        span_setter: SpanSetterArg = {"ents": True},
    ):
        super().__init__(nlp=nlp, name=name, span_setter=span_setter)

        if terms is None and regex is None:
            raise ValueError(
                "You must provide either `terms` or `regex` to the matcher."
            )

        terms = terms or {}
        regex = regex or {}

        self.attr = attr

        if term_matcher == "exact":
            self.phrase_matcher = EDSPhraseMatcher(
                nlp.vocab,
                attr=attr,
                ignore_excluded=ignore_excluded,
                ignore_space_tokens=ignore_space_tokens,
                **term_matcher_config,
            )
        elif term_matcher == "simstring":
            self.phrase_matcher = SimstringMatcher(
                nlp.vocab,
                attr=attr,
                ignore_excluded=ignore_excluded,
                ignore_space_tokens=ignore_space_tokens,
                **term_matcher_config,
            )
        else:
            raise ValueError(
                f"Algorithm {repr(term_matcher)} does not belong to"
                f" known matcher [exact, simstring]."
            )

        self.regex_matcher = RegexMatcher(
            attr=attr,
            ignore_excluded=ignore_excluded,
            ignore_space_tokens=ignore_space_tokens,
        )

        self.phrase_matcher.build_patterns(nlp=nlp, terms=terms)
        self.regex_matcher.build_patterns(regex=regex)

        self.set_extensions()

    def process(self, doc: Doc) -> List[Span]:
        """
        Find matching spans in doc.

        Parameters
        ----------
        doc:
            spaCy Doc object.

        Returns
        -------
        spans:
            List of Spans returned by the matchers.
        """

        matches = self.phrase_matcher(doc, as_spans=True)
        regex_matches = self.regex_matcher(doc, as_spans=True)

        spans = list(matches) + list(regex_matches)

        return spans

    def __call__(self, doc: Doc) -> Doc:
        """
        Adds spans to document.

        Parameters
        ----------
        doc:
            spaCy Doc object

        Returns
        -------
        doc:
            spaCy Doc object, annotated for extracted terms.
        """
        matches = self.process(doc)

        self.set_spans(doc, matches)

        return doc



main_pattern = dict(
    source="main",
    regex=[
        r"glomerulonephrite",
        r"(?<!pyelo)nephrite.{1,10}chronique",
        r"glomerulopathie",
        r"\bGNIgA",
        r"syndrome.{1,5}nephrotique",
        r"nephroangiosclerose",
        r"mal.de.bright",
        r"(maladie|syndrome).{1,7}berger",
        r"(maladie|syndrome).{1,7}bright",
        r"rachitisme.{1,5}renal",
        r"sydrome.{1,5}alport",
        r"good.?pasture",
        r"siadh",
        r"tubulopathie",
    ],
    exclude=dict(
        regex=[
            "aigu",
        ],
        window=4,
    ),
    regex_attr="NORM",
)

transplantation = dict(
    source="transplantation",
    regex=[
        r"transplant.{1,15}(rein|renal)",
        r"greff.{1,10}(rein|renal)",
    ],
    regex_attr="NORM",
)

acute_on_chronic = dict(
    source="acute_on_chronic",
    regex=[
        r"insuffisan.{1,10}(rein|renal).{1,5}aig.{1,10}chron",
    ],
    regex_attr="NORM",
)

dialysis = dict(
    source="dialysis",
    regex=[
        r"\beer\b",
        r"epuration extra.*renale",
        r"dialys",
    ],
    regex_attr="NORM",
    assign=[
        dict(
            name="chronic",
            regex=r"("
            + r"|".join(
                [
                    "long",
                    "chronique",
                    "peritoneal",
                    "depuis",
                    "intermitten",
                    "quotidien",
                    "hebdo",
                    "seances",
                    "reprise",
                    "poursuite",
                    "programme",
                    r"\blun",
                    r"\bmar",
                    r"\bmer",
                    r"\bjeu",
                    r"\bven",
                    r"\bsam",
                    r"\bdim",
                ]
            )
            + r")",
            window=(-5, 5),
        ),
    ],
)

general = dict(
    source="general",
    regex=[
        r"(insuffisan|fonction|malad).{1,10}\b(rein|rena)",
        r"\bmrc[^a-z]",
        r"\birc[^a-z]",
        r"nephropathie",
    ],
    regex_attr="NORM",
    assign=[
        dict(
            name="stage",
            regex=r"\b(iii|iv|v|3|4|5)\b",
            window=7,
            reduce_mode="keep_first",
        ),
        dict(
            name="status",
            regex=r"\b(moder|sever|terminal|pre.?greffe|post.?greffe|dialys|pre.?terminal)",  # noqa
            window=7,
            reduce_mode="keep_first",
        ),
        dict(
            name="dfg",
            regex=r"(?:dfg|debit.{1,10}filtration.{1,5}glomerulaire).*?(\d+[\.,]?\d+)",
            window=20,
            reduce_mode="keep_first",
        ),
    ],
)

acronym = dict(
    source="acronym",
    regex=[
        r"\bDPCA\b",
        r"\bGNMP\b",
        r"\bGEM\b",
        r"\bNCM\b",
    ],
    regex_attr="TEXT",
)

default_patterns = [
    main_pattern,
    transplantation,
    dialysis,
    general,
    acronym,
    acute_on_chronic,
]



def test_pipelines(doc):
    assert len(doc.ents) == 3
    patient, _, anomalie = doc.ents

    assert not patient._.negation
    assert anomalie._.negation

    assert not doc[0]._.history


def test_import_all():
    import edsnlp.pipes

    for name in dir(edsnlp.pipes):
        if not name.startswith("_"):
            getattr(edsnlp.pipes, name)



from datetime import timedelta
from typing import List, Optional, Set, Union

import pendulum
from loguru import logger
from spacy.tokens import Doc, Span, Token

from edsnlp.core import PipelineProtocol
from edsnlp.pipes.base import SpanGetterArg, get_spans
from edsnlp.pipes.qualifiers.base import RuleBasedQualifier
from edsnlp.pipes.terminations import termination as default_termination
from edsnlp.utils.deprecation import deprecated_getter_factory
from edsnlp.utils.filter import consume_spans, filter_spans
from edsnlp.utils.inclusion import check_inclusion, check_sent_inclusion

from . import patterns
from .patterns import sections_history


def history_getter(token: Union[Token, Span]) -> Optional[str]:
    if token._.history is True:
        return "ATCD"
    elif token._.history is False:
        return "CURRENT"
    else:
        return None


class HistoryQualifier(RuleBasedQualifier):
    """
    The `eds.history` pipeline uses a simple rule-based algorithm to detect spans that
    describe medical history rather than the diagnostic of a given visit.

    The mere definition of a medical history is not straightforward.
    Hence, this component only tags entities that are _explicitly described as part of
    the medical history_, e.g., preceded by a synonym of "medical history".

    This component may also use the output of:

    - the `eds.sections` component In that case, the
    entire `antécédent` section is tagged as a medical history.

    !!! warning "Sections"

        Be careful, the `eds.sections` component may oversize the `antécédents` section.
        Indeed, it detects *section titles* and tags the entire text between a title and
        the next as a section. Hence, should a section title goes undetected after the
        `antécédents` title, some parts of the document will erroneously be tagged as
        a medical history.

        To curb that possibility, using the output of the `eds.sections` component is
        deactivated by default.

    - the `eds.dates` component. In that case, it will take the
      dates into account to tag extracted entities as a medical history or not.

    !!! info "Dates"

        To take the most of the `eds.dates` component, you may add the ``note_datetime``
        context (cf. [Adding context][using-eds-nlps-helper-functions]). It allows the
        component to compute the duration of absolute dates
        (e.g., le 28 août 2022/August 28, 2022). The ``birth_datetime`` context allows
        the component to exclude the birthdate from the extracted dates.

    Examples
    --------
    The following snippet matches a simple terminology, and checks whether the extracted
    entities are history or not. It is complete and can be run _as is_.

    ```python
    import edsnlp, edsnlp.pipes as eds

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(eds.sentences())
    nlp.add_pipe(eds.normalizer())
    nlp.add_pipe(eds.sections())
    nlp.add_pipe(eds.dates())
    nlp.add_pipe(eds.matcher(terms=dict(douleur="douleur", malaise="malaises")))
    nlp.add_pipe(
        eds.history(
            use_sections=True,
            use_dates=True,
        ),
    )

    text = (
        "Le patient est admis le 23 août 2021 pour une douleur au bras. "
        "Il a des antécédents de malaises."
        "ANTÉCÉDENTS : "
        "- le patient a déjà eu des malaises. "
        "- le patient a eu une douleur à la jambe il y a 10 jours"
    )

    doc = nlp(text)

    doc.ents
    # Out: (douleur, malaises, malaises, douleur)

    doc.ents[0]._.history
    # Out: False

    doc.ents[1]._.history
    # Out: True

    doc.ents[2]._.history  # (1)
    # Out: True

    doc.ents[3]._.history  # (2)
    # Out: False
    ```

    1. The entity is in the section `antécédent`.
    2. The entity is in the section `antécédent`, however the extracted `relative_date`
    refers to an event that took place within 14 days.

    Extensions
    ----------
    The `eds.history` component declares two extensions, on both `Span` and `Token`
    objects :

    1. The `history` attribute is a boolean, set to `True` if the component predicts
       that the span/token is a medical history.
    2. The `history_` property is a human-readable string, computed from the `history`
       attribute. It implements a simple getter function that outputs `CURRENT` or
       `ATCD`, depending on the value of `history`.

    Parameters
    ----------
    nlp : PipelineProtocol
        The pipeline object.
    name : Optional[str]
        The component name.
    history : Optional[List[str]]
        List of terms indicating medical history reference.
    termination : Optional[List[str]]
        List of syntagms termination terms.
    use_sections : bool
        Whether to use section pipeline to detect medical history section.
    use_dates : bool
        Whether to use dates pipeline to detect if the event occurs
         a long time before the document date.
    attr : str
        spaCy's attribute to use:
        a string with the value "TEXT" or "NORM", or a dict with the key 'term_attr'
        we can also add a key for each regex.
    history_limit : Union[int, timedelta]
        The number of days after which the event is considered as history.
    exclude_birthdate : bool
        Whether to exclude the birthdate from history dates.
    closest_dates_only : bool
        Whether to include the closest dates only.
    span_getter : SpanGetterArg
        Which entities should be classified. By default, `doc.ents`
    on_ents_only : Union[bool, str, List[str], Set[str]]
        Deprecated, use `span_getter` instead.

        Whether to look for matches around detected entities only.
        Useful for faster inference in downstream tasks.

        - If True, will look in all ents located in `doc.ents` only
        - If an iterable of string is passed, will additionally look in `doc.spans[key]`
        for each key in the iterable
    explain : bool
        Whether to keep track of cues for each entity.

    Authors and citation
    --------------------
    The `eds.history` component was developed by AP-HP's Data Science team.
    """

    history_limit: timedelta

    def __init__(
        self,
        nlp: PipelineProtocol,
        name: Optional[str] = "history",
        *,
        history: Optional[List[str]] = None,
        termination: Optional[List[str]] = None,
        use_sections: bool = False,
        use_dates: bool = False,
        attr: str = "NORM",
        history_limit: int = 14,
        closest_dates_only: bool = True,
        exclude_birthdate: bool = True,
        span_getter: Optional[SpanGetterArg] = None,
        on_ents_only: Optional[Union[bool, str, List[str], Set[str]]] = None,
        explain: bool = False,
    ):
        terms = dict(
            history=patterns.history if history is None else history,
            termination=default_termination if termination is None else termination,
        )

        super().__init__(
            nlp=nlp,
            name=name,
            attr=attr,
            explain=explain,
            terms=terms,
            on_ents_only=on_ents_only,
            span_getter=span_getter,
        )

        self.history_limit = timedelta(history_limit)
        self.exclude_birthdate = exclude_birthdate
        self.closest_dates_only = closest_dates_only

        self.sections = use_sections and (
            "eds.sections" in nlp.pipe_names or "sections" in nlp.pipe_names
        )
        if use_sections and not self.sections:
            logger.warning(
                "You have requested that the pipeline use annotations "
                "provided by the `section` pipeline, but it was not set. "
                "Skipping that step."
            )

        self.dates = use_dates and (
            "eds.dates" in nlp.pipe_names or "dates" in nlp.pipe_names
        )
        if use_dates:
            if not self.dates:
                logger.warning(
                    "You have requested that the pipeline use dates "
                    "provided by the `dates` pipeline, but it was not set. "
                    "Skipping that step."
                )
            elif exclude_birthdate:
                logger.info(
                    "You have requested that the pipeline use date "
                    "and exclude birth dates. "
                    "To make the most of this feature, "
                    "make sur you provide the `birth_datetime` "
                    "context and `note_datetime` context. "
                )
            else:
                logger.info(
                    "You have requested that the pipeline use date "
                    "To make the most of this feature, "
                    "make sure you provide the `note_datetime` "
                    "context. "
                )

    def set_extensions(self) -> None:
        for cls in (Token, Span):
            if not cls.has_extension("history"):
                cls.set_extension("history", default=None)

            if not cls.has_extension("antecedents"):
                cls.set_extension(
                    "antecedents",
                    getter=deprecated_getter_factory("antecedents", "history"),
                )

            if not cls.has_extension("antecedent"):
                cls.set_extension(
                    "antecedent",
                    getter=deprecated_getter_factory("antecedent", "history"),
                )

            if not cls.has_extension("history_"):
                cls.set_extension("history_", getter=history_getter)

            if not cls.has_extension("antecedents_"):
                cls.set_extension(
                    "antecedents_",
                    getter=deprecated_getter_factory("antecedents_", "history_"),
                )

            if not cls.has_extension("antecedent_"):
                cls.set_extension(
                    "antecedent_",
                    getter=deprecated_getter_factory("antecedent_", "history_"),
                )

        # Store history mentions responsible for the history entity's character
        if not Span.has_extension("history_cues"):
            Span.set_extension("history_cues", default=[])

        # Store recent mentions responsible for the non-antecedent entity's character
        if not Span.has_extension("recent_cues"):
            Span.set_extension("recent_cues", default=[])

        if not Span.has_extension("antecedents_cues"):
            Span.set_extension(
                "antecedents_cues",
                getter=deprecated_getter_factory("antecedents_cues", "history_cues"),
            )

        if not Span.has_extension("antecedent_cues"):
            Span.set_extension(
                "antecedent_cues",
                getter=deprecated_getter_factory("antecedent_cues", "history_cues"),
            )

    def process(self, doc: Doc) -> Doc:
        note_datetime = None
        if doc._.note_datetime is not None:
            try:
                note_datetime = pendulum.instance(doc._.note_datetime)
                note_datetime = note_datetime.set(tz="Europe/Paris")
            except ValueError:
                logger.debug(
                    "note_datetime must be a datetime objects. "
                    "Skipping history qualification from note_datetime."
                )
                note_datetime = None

        birth_datetime = None
        if doc._.birth_datetime is not None:
            try:
                birth_datetime = pendulum.instance(doc._.birth_datetime)
                birth_datetime = birth_datetime.set(tz="Europe/Paris")
            except ValueError:
                logger.debug(
                    "birth_datetime must be a datetime objects. "
                    "Skipping history qualification from birth date."
                )
                birth_datetime = None

        matches = self.get_matches(doc)

        terminations = [m for m in matches if m.label_ == "termination"]
        boundaries = self._boundaries(doc, terminations)

        # Removes duplicate matches and pseudo-expressions in one statement
        matches = filter_spans(matches, label_to_remove="pseudo")

        entities = list(get_spans(doc, self.span_getter))
        ents = None
        sub_sections = None
        sub_recent_dates = None
        sub_history_dates = None

        sections = []
        if self.sections:
            sections = [
                Span(doc, section.start, section.end, label="ATCD")
                for section in doc.spans["sections"]
                if section.label_ in sections_history
            ]

        history_dates = []
        recent_dates = []
        if self.dates:
            for date in doc.spans["dates"]:
                value = date._.date
                if value.mode == "relative":
                    if value.direction.value == "current":
                        if (
                            (value.year == 0 and self.history_limit >= timedelta(365))
                            or (
                                value.month == 0 and self.history_limit >= timedelta(30)
                            )
                            or (value.week == 0 and self.history_limit >= timedelta(7))
                            or (value.day == 0)
                        ):
                            recent_dates.append(
                                Span(doc, date.start, date.end, label="relative_date")
                            )
                    elif value.direction.value == "past":
                        if (
                            -value.to_duration(
                                note_datetime=doc._.note_datetime,
                                infer_from_context=True,
                                tz="Europe/Paris",
                                default_day=15,
                            )
                            >= self.history_limit
                        ):
                            history_dates.append(
                                Span(doc, date.start, date.end, label="relative_date")
                            )
                        else:
                            recent_dates.append(
                                Span(doc, date.start, date.end, label="relative_date")
                            )
                elif value.mode == "absolute" and doc._.note_datetime:
                    try:
                        absolute_date = value.to_datetime(
                            note_datetime=note_datetime,
                            infer_from_context=True,
                            tz="Europe/Paris",
                            default_day=15,
                        )
                    except ValueError as e:
                        absolute_date = None
                        logger.warning(
                            "In doc {}, the following date {} raises this error: {}. "
                            "Skipping this date.",
                            doc._.note_id,
                            value,
                            e,
                        )
                    if absolute_date:
                        if note_datetime.diff(absolute_date) < self.history_limit:
                            recent_dates.append(
                                Span(doc, date.start, date.end, label="absolute_date")
                            )
                        elif not (
                            self.exclude_birthdate
                            and birth_datetime
                            and absolute_date == birth_datetime
                        ):
                            history_dates.append(
                                Span(doc, date.start, date.end, label="absolute_date")
                            )

        for start, end in boundaries:
            ents, entities = consume_spans(
                entities,
                filter=lambda s: check_inclusion(s, start, end),
                second_chance=ents,
            )

            sub_matches, matches = consume_spans(
                matches, lambda s: start <= s.start < end
            )

            if self.sections:
                sub_sections, sections = consume_spans(
                    sections, lambda s: s.start < end <= s.end, sub_sections
                )

            close_recent_dates = []
            close_history_dates = []
            if self.dates:
                sub_recent_dates, recent_dates = consume_spans(
                    recent_dates,
                    lambda s: check_sent_inclusion(s, start, end),
                    sub_recent_dates,
                )
                sub_history_dates, history_dates = consume_spans(
                    history_dates,
                    lambda s: check_sent_inclusion(s, start, end),
                    sub_history_dates,
                )

                # Filter dates inside the boundaries only
                if self.closest_dates_only:
                    if sub_recent_dates:
                        close_recent_dates = [
                            recent_date
                            for recent_date in sub_recent_dates
                            if check_inclusion(recent_date, start, end)
                        ]
                        if sub_history_dates:
                            close_history_dates = [
                                history_date
                                for history_date in sub_history_dates
                                if check_inclusion(history_date, start, end)
                            ]
                            # If no date inside the boundaries, get the closest
                            if not close_recent_dates and not close_history_dates:
                                min_distance_recent_date = min(
                                    [
                                        abs(sub_recent_date.start - start)
                                        for sub_recent_date in sub_recent_dates
                                    ]
                                )
                                min_distance_history_date = min(
                                    [
                                        abs(sub_history_date.start - start)
                                        for sub_history_date in sub_history_dates
                                    ]
                                )
                                if min_distance_recent_date < min_distance_history_date:
                                    close_recent_dates = [
                                        min(
                                            sub_recent_dates,
                                            key=lambda x: abs(x.start - start),
                                        )
                                    ]
                                else:
                                    close_history_dates = [
                                        min(
                                            sub_history_dates,
                                            key=lambda x: abs(x.start - start),
                                        )
                                    ]
                        elif not close_recent_dates:
                            close_recent_dates = [
                                min(
                                    sub_recent_dates,
                                    key=lambda x: abs(x.start - start),
                                )
                            ]
                    elif sub_history_dates:
                        close_history_dates = [
                            history_date
                            for history_date in sub_history_dates
                            if check_inclusion(history_date, start, end)
                        ]
                        # If no date inside the boundaries, get the closest
                        if not close_history_dates:
                            close_history_dates = [
                                min(
                                    sub_history_dates,
                                    key=lambda x: abs(x.start - start),
                                )
                            ]

            if self.on_ents_only and not ents:
                continue

            history_cues = [m for m in sub_matches if m.label_ == "history"]
            recent_cues = []

            if self.sections:
                history_cues.extend(sub_sections)

            if self.dates:
                history_cues.extend(
                    close_history_dates
                    if self.closest_dates_only
                    else sub_history_dates
                )
                recent_cues.extend(
                    close_recent_dates if self.closest_dates_only else sub_recent_dates
                )

            history = bool(history_cues) and not bool(recent_cues)

            for ent in ents:
                ent._.history = ent._.history or history

                if self.explain:
                    ent._.history_cues += history_cues
                    ent._.recent_cues += recent_cues
                if not self.on_ents_only:
                    for token in ent:
                        token._.history = token._.history or history

        return doc



import re
import shutil
import tempfile
from textwrap import dedent
from typing import Tuple

from markdown.extensions import Extension
from markdown.extensions.attr_list import get_attrs
from markdown.extensions.codehilite import parse_hl_lines
from markdown.extensions.fenced_code import FencedBlockPreprocessor
from mkdocs.commands.build import build
from mkdocs.config import load_config
from mkdocs.config.config_options import Type as MkType
from mkdocs.config.defaults import MkDocsConfig
from mkdocs.plugins import BasePlugin
from mkdocstrings.extension import AutoDocProcessor
from mkdocstrings.plugin import MkdocstringsPlugin

BRACKET_RE = re.compile(r"\[([^\[]+)\]")
CITE_RE = re.compile(r"@([\w_:-]+)")
DEF_RE = re.compile(r"\A {0,3}\[@([\w_:-]+)\]:\s*(.*)")
INDENT_RE = re.compile(r"\A\t| {4}(.*)")

CITATION_RE = r"(\[@(?:[\w_:-]+)(?: *, *@(?:[\w_:-]+))*\])"


class PyCodePreprocessor(FencedBlockPreprocessor):
    """Gather reference definitions and citation keys"""

    FENCED_BLOCK_RE = re.compile(
        dedent(
            r"""
            (?P<fence>^[ ]*(?:~{3,}|`{3,}))[ ]*                          # opening fence
            ((\{(?P<attrs>[^\}\n]*)\})|                              # (optional {attrs} or
            (\.?(?P<lang>[\w#.+-]*)[ ]*)?                            # optional (.)lang
            (hl_lines=(?P<quot>"|')(?P<hl_lines>.*?)(?P=quot)[ ]*)?) # optional hl_lines)
            \n                                                       # newline (end of opening fence)
            (?P<code>.*?)(?<=\n)                                     # the code block
            (?P=fence)[ ]*$                                          # closing fence
        """  # noqa: E501
        ),
        re.MULTILINE | re.DOTALL | re.VERBOSE,
    )

    def __init__(self, md, code_blocks):
        super().__init__(md, {})
        self.code_blocks = code_blocks

    def run(self, lines):
        text = "\n".join(lines)
        while True:
            # ----  https://github.com/Python-Markdown/markdown/blob/5a2fee/markdown/extensions/fenced_code.py#L84C9-L98  # noqa: E501
            m = self.FENCED_BLOCK_RE.search(text)
            if m:
                lang, id, classes, config = None, "", [], {}
                if m.group("attrs"):
                    id, classes, config = self.handle_attrs(get_attrs(m.group("attrs")))
                    if len(classes):
                        lang = classes.pop(0)
                else:
                    if m.group("lang"):
                        lang = m.group("lang")
                    if m.group("hl_lines"):
                        # Support `hl_lines` outside of `attrs` for
                        # backward-compatibility
                        config["hl_lines"] = parse_hl_lines(m.group("hl_lines"))
                # ----
                code = m.group("code")

                if lang == "python" and "no-check" not in classes:
                    self.code_blocks.append(dedent(code))
            else:
                break
            text = text[m.end() :]

        return lines


context_citations = None


class PyCodeExtension(Extension):
    def __init__(self, code_blocks):
        super(PyCodeExtension, self).__init__()
        self.code_blocks = code_blocks

    def extendMarkdown(self, md):
        self.md = md
        md.registerExtension(self)
        md.preprocessors.register(
            PyCodePreprocessor(md, self.code_blocks), "fenced_code", 31
        )
        for ext in md.registeredExtensions:
            if isinstance(ext, AutoDocProcessor):
                ext._config["mdx"].append(self)


def makeExtension(*args, **kwargs):
    return PyCodeExtension(*args, **kwargs)


class PyCodeExtractorPlugin(BasePlugin):
    config_scheme: Tuple[Tuple[str, MkType]] = (
        # ("bibtex_file", MkType(str)),  # type: ignore[assignment]
        # ("order", MkType(str, default="unsorted")),  # type: ignore[assignment]
    )

    def __init__(self, global_config):
        self.global_config = global_config
        self.page_code_blocks = []
        self.docs_code_blocks = []

    def on_config(self, config: MkDocsConfig):
        self.ext = PyCodeExtension(self.page_code_blocks)
        # After pymdownx.highlight, because of weird registering deleting the first
        # extension
        config["markdown_extensions"].append(self.ext)
        config["markdown_extensions"].remove("pymdownx.highlight")
        config["markdown_extensions"].remove("fenced_code")

    def on_pre_build(self, *, config: MkDocsConfig):
        mkdocstrings_plugin: MkdocstringsPlugin = config.plugins["mkdocstrings"]
        mkdocstrings_plugin.get_handler("python")

    def on_page_content(self, html, page, config, files):
        if len(self.page_code_blocks):
            self.docs_code_blocks.append((page.url, "\n".join(self.page_code_blocks)))
        self.page_code_blocks.clear()
        return html


def extract_docs_code():
    config = load_config()

    temp_dir = tempfile.mkdtemp()
    try:
        config["site_dir"] = temp_dir

        # plug the pycode extractor plugin
        plugin = PyCodeExtractorPlugin(config)
        config.plugins["pycode_extractor"] = plugin

        config["plugins"].run_event("startup", command="build", dirty=False)
        try:
            build(config)
        finally:
            config["plugins"].run_event("shutdown")

    finally:
        shutil.rmtree(temp_dir, ignore_errors=True)

    docs_code_blocks = plugin.docs_code_blocks
    # Deduplicate both keys and values
    return {
        k: v
        for v, k in {v: k for k, v in docs_code_blocks if "changelog" not in k}.items()
    }



results_congestive_heart_failure = dict(
    has_match=[
        True,
        True,
        False,
        True,
        False,
    ],
    detailled_status="PRESENT",
    assign=None,
    texts=[
        "Présence d'un oedème pulmonaire",
        "Le patient est équipé d'un pace-maker",
        "Un cardiopathie non décompensée",  # no decompensation
        "Insuffisance cardiaque",
        "Insuffisance cardiaque minime",  # minimal severity
    ],
)



from edsnlp.core import registry

from .history import HistoryQualifier

DEFAULT_CONFIG = dict(
    history=None,
    termination=None,
    use_sections=False,
    use_dates=False,
    attr="NORM",
    history_limit=14,
    closest_dates_only=True,
    exclude_birthdate=True,
    span_getter=None,
    on_ents_only=True,
    explain=False,
)

create_component = registry.factory.register(
    "eds.history",
    assigns=["span._.history"],
    deprecated=[
        "history",
        "antecedents",
        "eds.antecedents",
    ],
)(HistoryQualifier)



from functools import partial
from typing import Any, Iterable, List, TypeVar

from edsnlp.utils.collections import batchify
from edsnlp.utils.span_getters import get_spans

T = TypeVar("T")


def batchify_by_words(
    iterable: Iterable[T],
    batch_size: int,
    drop_last: bool = False,
) -> Iterable[List[T]]:
    """
    Yields batch that contain at most `batch_size` words.
    If an item contains more than `batch_size` words, it will be yielded as a single
    batch.

    Parameters
    ----------
    iterable
    batch_size
    drop_last
    """
    batch = []
    total = 0
    for item in iterable:
        count = len(item)
        if total + count > batch_size and total > 0:
            yield batch
            batch = []
            total = 0
        batch.append(item)
        total += count
    if len(batch) > 0 and not drop_last:
        yield batch


def batchify_by_padded_words(
    iterable: Iterable[T],
    batch_size: int,
    drop_last: bool = False,
) -> Iterable[List[T]]:
    batch = []
    max_words = 0
    for item in iterable:
        count = len(item)
        if (1 + len(batch)) * max(max_words, count) > batch_size:
            yield batch
            batch = []
            max_words = 0
        batch.append(item)
        max_words = max(max_words, count)
    if len(batch) > 0 and not drop_last:
        yield batch


def batchify_by_spans(
    iterable: Iterable[T],
    batch_size: int,
    span_getter: Any = {"ents": True},
    drop_last: bool = False,
) -> Iterable[List[T]]:
    batch = []
    total = 0
    for item in iterable:
        count = len(list(get_spans(item, span_getter)))
        if total + count > batch_size:
            yield batch
            batch = []
            total = 0
        batch.append(item)
        total += count
    if len(batch) > 0 and not drop_last:
        yield batch


def batchify_with_counts(
    iterable,
    batch_size,
):
    total = 0
    batch = []
    for item, count in iterable:
        if total + count > batch_size:
            yield batch, total
            batch = []
            total = 0
        batch.append(item)
        total += count
    if len(batch) > 0:
        yield batch, total


batchify_fns = {
    "words": batchify_by_words,
    "padded_words": batchify_by_padded_words,
    "spans": batchify_by_spans,
    "docs": batchify,
    "ents": partial(batchify_by_spans, span_getter={"ents": True}),
}



from edsnlp.utils.regex import make_pattern

raw_delimiters = [r"\/", r"[-−]"]
delimiters = raw_delimiters + [r"\.", r"[^\S]+"]

raw_delimiter_pattern = make_pattern(raw_delimiters)
raw_delimiter_with_spaces_pattern = make_pattern(raw_delimiters + [r"[^\S]+"])
delimiter_pattern = make_pattern(delimiters)

ante_num_pattern = (
    f"(?<!.(?:{raw_delimiter_pattern})|[.:%a-zA-Z]|[0-9][.:%][ ]?|[0-9][,]?)"
)
post_num_pattern = f"(?!{raw_delimiter_pattern}|[%a-zA-Z]|[ ]?[.:%][0-9]|[.,:]?[0-9])"

ante_num_with_letter_pattern = "(?<!/|[.:%]|[0-9][-−.:%][ ]?|[0-9][,]?)"
post_num_with_letter_pattern = "(?!/|[%]|[ ]?[-−.:%][0-9]|[.,:]?[0-9])"



from typing import Any, Callable, Iterable, List, Optional, Sequence, Tuple, Union

from spacy.tokens import Span
from spacy.tokens.doc import Doc


def default_sort_key(span: Span) -> Tuple[int, int]:
    """
    Returns the sort key for filtering spans.

    Parameters
    ----------
    span : Span
        Span to sort.

    Returns
    -------
    key : Tuple(int, int)
        Sort key.
    """
    if isinstance(span, tuple):
        span = span[0]
    return span.end - span.start, -span.start


def start_sort_key(span: Union[Span, Tuple[Span, Any]]) -> Tuple[int, int]:
    """
    Returns the sort key for filtering spans by start order.

    Parameters
    ----------
    span : Span
        Span to sort.

    Returns
    -------
    key : Tuple(int, int)
        Sort key.
    """
    if isinstance(span, tuple):
        span = span[0]
    return span.start


def filter_spans(
    spans: Iterable[Union["Span", Tuple["Span", Any]]],
    label_to_remove: Optional[str] = None,
    return_discarded: bool = False,
    sort_key: Callable[[Span], Any] = default_sort_key,
) -> Union[
    List[Union[Span, Tuple[Span, Any]]],
    Tuple[List[Union[Span, Tuple[Span, Any]]], List[Union[Span, Tuple[Span, Any]]]],
]:
    """
    Re-definition of spacy's filtering function, that returns discarded spans
    as well as filtered ones.

    Can also accept a `label_to_remove` argument, useful for filtering out
    pseudo cues. If set, `results` can contain overlapping spans: only
    spans overlapping with excluded labels are removed. The main expected
    use case is for pseudo-cues.

    It can handle an iterable of tuples instead of an iterable of `Span`s.
    The primary use-case is the use with the `RegexMatcher`'s capacity to
    return the span's `groupdict`.

    !!! note ""

        The **spaCy documentation states**:

        > Filter a sequence of spans and remove duplicates or overlaps.
        > Useful for creating named entities (where one token can only
        > be part of one entity) or when merging spans with
        > `Retokenizer.merge`. When spans overlap, the (first)
        > longest span is preferred over shorter spans.

    !!! danger "Filtering out spans"

        If the `label_to_remove` argument is supplied, it might be tempting to
        filter overlapping spans that are not part of a label to remove.

        The reason we keep all other possibly overlapping labels is that in qualifier
        pipelines, the same cue can precede **and** follow a marked entity.
        Hence we need to keep every example.

    Parameters
    ----------
    spans : Iterable[Union["Span", Tuple["Span", Any]]]
        Spans to filter.
    return_discarded : bool
        Whether to return discarded spans.
    label_to_remove : str, optional
        Label to remove. If set, results can contain overlapping spans.
    sort_key : Callable[Span, Any], optional
        Key to sorting spans before applying overlap conflict resolution.
        A span with a higher key will have precedence over another span.
        By default, the largest, leftmost spans are selected first.

    Returns
    -------
    results : List[Union[Span, Tuple[Span, Any]]]
        Filtered spans
    discarded : List[Union[Span, Tuple[Span, Any]]], optional
        Discarded spans
    """
    sorted_spans = sorted(spans, key=sort_key, reverse=True)
    result = []
    discarded = []
    seen_tokens = set()
    for span in sorted_spans:
        s = span if isinstance(span, Span) else span[0]
        # Check for end - 1 here because boundaries are inclusive
        token_range = set(range(s.start, s.end))
        if token_range.isdisjoint(seen_tokens):
            if label_to_remove is None or s.label_ != label_to_remove:
                result.append(span)
            if label_to_remove is None or s.label_ == label_to_remove:
                seen_tokens.update(token_range)
        elif label_to_remove is None or s.label_ != label_to_remove:
            discarded.append(span)

    result = sorted(result, key=start_sort_key)
    discarded = sorted(discarded, key=start_sort_key)

    if return_discarded:
        return result, discarded

    return result


def consume_spans(
    spans: List[Span],
    filter: Callable,
    second_chance: Optional[List[Span]] = None,
) -> Tuple[List[Span], List[Span]]:
    """
    Consume a list of span, according to a filter.

    !!! warning
        This method makes the hard hypothesis that:

        1. Spans are sorted.
        2. Spans are consumed in sequence and only once.

        The second item is problematic for the way we treat long entities,
        hence the `second_chance` parameter, which lets entities be seen
        more than once.

    Parameters
    ----------
    spans : List of spans
        List of spans to filter
    filter : Callable
        Filtering function. Should return True when the item is to be included.
    second_chance : List of spans, optional
        Optional list of spans to include again (useful for long entities),
        by default None

    Returns
    -------
    matches : List of spans
        List of spans consumed by the filter.
    remainder : List of spans
        List of remaining spans in the original `spans` parameter.
    """

    if not second_chance:
        second_chance = []
    else:
        second_chance = [m for m in second_chance if filter(m)]

    if not spans:
        return second_chance, []

    for i, span in enumerate(spans):
        if not filter(span):
            break
        else:
            i += 1

    matches = spans[:i]
    remainder = spans[i:]

    matches.extend(second_chance)

    return matches, remainder


def get_spans(spans: List[Span], label: Union[int, str]) -> List[Span]:
    """
    Extracts spans with a given label.
    Prefer using hash label for performance reasons.

    Parameters
    ----------
    spans : List[Span]
        List of spans to filter.
    label : Union[int, str]
        Label to filter on.

    Returns
    -------
    List[Span]
        Filtered spans.
    """
    if isinstance(label, int):
        return [span for span in spans if span.label == label]
    else:
        return [span for span in spans if span.label_ == label]


def span_f1(a: Span, b: Span) -> float:
    """
    Computes the F1 overlap between two spans.

    Parameters
    ----------
    a: Span
        First span
    b: Span
        Second span

    Returns
    -------
    float
        F1 overlap
    """
    start_a, end_a = a.start, a.end
    start_b, end_b = b.start, b.end
    overlap = max(0, min(end_a, end_b) - max(start_a, start_b))
    return 2 * overlap / (end_a - start_a + end_b - start_b)


def align_spans(
    source_spans: Sequence[Span],
    target_spans: Sequence[Span],
    sort_by_overlap: bool = False,
) -> List[List[Span]]:
    """
    Aligns two lists of spans, by matching source spans that overlap target spans.
    This function is optimized to avoid quadratic complexity.

    Parameters
    ----------
    source_spans : List[Span]
        List of spans to align.
    target_spans : List[Span]
        List of spans to align.
    sort_by_overlap : bool
        Whether to sort the aligned spans by maximum dice/f1 overlap
        with the target span.

    Returns
    -------
    List[List[Span]]
        Subset of `source` spans for each target span
    """
    source_spans = sorted(source_spans, key=lambda x: (x.start, x.end))
    targets_with_idx = sorted(
        enumerate(target_spans), key=lambda x: (x[1].start, x[1].end)
    )

    aligned = [set() for _ in target_spans]
    source_idx = 0
    for target_idx, target in targets_with_idx:
        while (
            source_idx < len(source_spans)
            and source_spans[source_idx].end <= target.start
        ):
            source_idx += 1
        i = source_idx
        while i < len(source_spans) and source_spans[i].start < target.end:
            aligned[target_idx].add(source_spans[i])
            i += 1

    aligned = [list(span_set) for span_set in aligned]

    # Sort the aligned spans by maximum dice/f1 overlap with the target span
    if sort_by_overlap:
        aligned = [
            sorted(
                aligned_span_set, key=lambda x: span_f1(x, target_span), reverse=True
            )
            for aligned_span_set, target_span in zip(aligned, target_spans)
        ]

    return aligned


def get_span_group(doclike: Union[Doc, Span], group: str) -> List[Span]:
    """
    Get the spans of a span group that are contained inside a doclike object.

    Parameters
    ----------
    doclike : Union[Doc, Span]
        Doclike object to act as a mask.
    group : str
        Group name from which to get the spans.

    Returns
    -------
    List[Span]
        List of spans.
    """
    if isinstance(doclike, Doc):
        return [span for span in doclike.spans.get(group, ())]
    else:
        return [
            span
            for span in doclike.doc.spans.get(group, ())
            if span.start >= doclike.start and span.end <= doclike.end
        ]



from typing import TYPE_CHECKING

from edsnlp import registry

from .span_linker import TrainableSpanLinker

create_component = registry.factory.register(
    "eds.span_linker",
    assigns=[],
    deprecated=[],
)(TrainableSpanLinker)

if TYPE_CHECKING:
    create_component = TrainableSpanLinker



import torch

IMPOSSIBLE = -100000000


def repeat_like(x, y):
    return x.repeat(tuple((a if b == 1 else 1) for a, b in zip(y.shape, x.shape)))


def masked_flip(x, mask, dim_x=-2):
    mask = repeat_like(mask, x)
    flipped_x = torch.zeros_like(x)
    flipped_x[mask] = x.flip(dim_x)[mask.flip(-1)]
    return flipped_x


@torch.jit.script
def logsumexp_reduce(log_A, log_B):
    # log_A: 2 * N * M
    # log_B: 2 *     M * O
    # out: 2 * N * O
    return (log_A.unsqueeze(-1) + log_B.unsqueeze(-3)).logsumexp(-2)


@torch.jit.script
def max_reduce(log_A, log_B):
    # log_A: 2 * N * M
    # log_B: 2 *     M * O
    # out: 2 * N * O
    return (log_A.unsqueeze(-1) + log_B.unsqueeze(-3)).max(-2)


def index_dim(X, Y, dim):
    ndims = X.dim()
    if dim < 0:
        dim += ndims
    assert (
        0 <= dim < ndims
    ), f"Index out of range: {dim} for tensor of {ndims} dimensions"
    index_dims = list(
        torch.meshgrid(
            *(torch.arange(size) for i, size in enumerate(X.shape) if i != dim)
        )
    )
    return X[(*index_dims[:dim], Y, *index_dims[dim:])]


# noinspection PyTypeChecker
class LinearChainCRF(torch.nn.Module):
    def __init__(
        self,
        forbidden_transitions,
        start_forbidden_transitions=None,
        end_forbidden_transitions=None,
        learnable_transitions=True,
        with_start_end_transitions=True,
    ):
        """
        A linear chain CRF in Pytorch

        Parameters
        ----------
        forbidden_transitions: torch.BoolTensor
            Shape: n_tags * n_tags
            Impossible transitions (1 means impossible) from position n to position n+1
        start_forbidden_transitions: Optional[torch.BoolTensor]
            Shape: n_tags
            Impossible transitions at the start of a sequence
        end_forbidden_transitions Optional[torch.BoolTensor]
            Shape: n_tags
            Impossible transitions at the end of a sequence
        learnable_transitions: bool
            Should we learn transition scores to complete the
            constraints ?
        with_start_end_transitions:
            Should we apply start-end transitions.
            If learnable_transitions is True, learn start/end transition scores
        """
        super().__init__()

        num_tags = forbidden_transitions.shape[0]

        self.with_start_end_transitions = with_start_end_transitions
        self.register_buffer("forbidden_transitions", forbidden_transitions.bool())
        self.register_buffer(
            "start_forbidden_transitions",
            start_forbidden_transitions.bool()
            if start_forbidden_transitions is not None
            else torch.zeros(num_tags, dtype=torch.bool),
        )
        self.register_buffer(
            "end_forbidden_transitions",
            end_forbidden_transitions.bool()
            if end_forbidden_transitions is not None
            else torch.zeros(num_tags, dtype=torch.bool),
        )
        if learnable_transitions:
            self.transitions = torch.nn.Parameter(
                torch.zeros_like(forbidden_transitions, dtype=torch.float)
            )
        else:
            self.register_buffer(
                "transitions",
                torch.zeros_like(forbidden_transitions, dtype=torch.float),
            )

        if learnable_transitions and with_start_end_transitions:
            self.start_transitions = torch.nn.Parameter(
                torch.zeros(num_tags, dtype=torch.float)
            )
        else:
            self.register_buffer(
                "start_transitions", torch.zeros(num_tags, dtype=torch.float)
            )

        if learnable_transitions and with_start_end_transitions:
            self.end_transitions = torch.nn.Parameter(
                torch.zeros(num_tags, dtype=torch.float)
            )
        else:
            self.register_buffer(
                "end_transitions", torch.zeros(num_tags, dtype=torch.float)
            )

    def decode(self, emissions, mask):
        """
        Decodes a sequence of tag scores using the Viterbi algorithm

        Parameters
        ----------
        emissions: torch.FloatTensor
            Shape: ... * n_tokens * n_tags
        mask: torch.BoolTensor
            Shape: ... * n_tokens

        Returns
        -------
        torch.LongTensor
            Backtrack indices (= argmax), ie best tag sequence
        """
        transitions = self.transitions.masked_fill(
            self.forbidden_transitions, IMPOSSIBLE
        )
        start_transitions = end_transitions = torch.zeros_like(self.start_transitions)
        if self.with_start_end_transitions:
            start_transitions = self.start_transitions.masked_fill(
                self.start_forbidden_transitions, IMPOSSIBLE
            )
            end_transitions = self.end_transitions.masked_fill(
                self.end_forbidden_transitions, IMPOSSIBLE
            )
        path = torch.zeros(*emissions.shape[:-1], dtype=torch.long)

        emissions[..., 1:][~mask] = IMPOSSIBLE
        emissions = emissions.unbind(1)  # 1 is axis for words

        # emissions: n_tokens * n_samples * n_tags
        out = [emissions[0] + start_transitions]
        backtrack = []

        for k in range(1, len(emissions)):
            res, indices = max_reduce(out[-1], transitions)
            backtrack.append(indices)
            out.append(res + emissions[k])

        res, indices = max_reduce(out[-1], end_transitions.unsqueeze(-1))
        path[:, -1] = indices.squeeze(-1)

        # If make has shape n_samples * n_tokens,
        # we only need range(n_samples)
        if len(backtrack) > 1:
            # Backward max path following
            for k, b in enumerate(backtrack[::-1]):
                path[:, -k - 2] = index_dim(b, path[:, -k - 1], dim=-1)

        return path.to(transitions.device)

    def marginal(self, emissions, mask):
        """
        Compute the marginal log-probabilities of the tags
        given the emissions and the transition probabilities and
        constraints of the CRF

        We could use the `propagate` method but this implementation
        is faster.

        Parameters
        ----------
        emissions: torch.FloatTensor
            Shape: ... * n_tokens * n_tags
        mask: torch.BoolTensor
            Shape: ... * n_tokens

        Returns
        -------
        torch.FloatTensor
            Shape: ... * n_tokens * n_tags
        """
        device = emissions.device

        transitions = self.transitions.masked_fill(
            self.forbidden_transitions, IMPOSSIBLE
        )
        start_transitions = end_transitions = torch.zeros_like(self.start_transitions)
        if self.with_start_end_transitions:
            start_transitions = self.start_transitions.masked_fill(
                self.start_forbidden_transitions, IMPOSSIBLE
            )
            end_transitions = self.end_transitions.masked_fill(
                self.end_forbidden_transitions, IMPOSSIBLE
            )

        bi_transitions = torch.stack([transitions, transitions.t()], dim=0).unsqueeze(1)

        # add start transitions (ie cannot start with ...)
        emissions[:, 0] = emissions[:, 0] + start_transitions

        # add end transitions (ie cannot end with ...)
        emissions[
            torch.arange(mask.shape[0], device=device), mask.long().sum(1) - 1
        ] = (
            emissions[
                torch.arange(mask.shape[0], device=device),
                mask.long().sum(1) - 1,
            ]
            + end_transitions
        )

        # stack start -> end emissions (need to flip the previously flipped emissions)
        # and end -> start emissions.
        # emissions: n_samples * n_tokens * ... * n_tags
        # bi_emissions: n_tokens * (2 * n_samples *  * ... * n_tags)
        bi_emissions = torch.stack(
            [emissions, masked_flip(emissions, mask, dim_x=1)], 0
        ).unbind(2)

        out = [bi_emissions[0]]
        for word_bi_emissions in bi_emissions[1:]:
            res = logsumexp_reduce(out[-1], bi_transitions)
            out.append(res + word_bi_emissions)

        # out shape: 2 * n_samples * n_tokens * ... * n_tags
        out = torch.stack(out, dim=2)

        forward = out[0]
        backward = masked_flip(out[1], mask, dim_x=1)
        backward_z = backward.logsumexp(-1)

        return forward + backward - emissions - backward_z[:, :, :, None]

    def forward(self, emissions, mask, target):
        """
        Compute the posterior reduced log-probabilities of the tags
        given the emissions and the transition probabilities and
        constraints of the CRF, ie the loss.


        We could use the `propagate` method but this implementation
        is faster.

        Parameters
        ----------
        emissions: torch.FloatTensor
            Shape: n_samples * n_tokens * ... * n_tags
        mask: torch.BoolTensor
            Shape: n_samples * n_tokens * ...
        target: torch.BoolTensor
            Shape: n_samples * n_tokens * ... * n_tags
            The target tags represented with 1-hot encoding
            We use 1-hot instead of long format to handle
            cases when multiple tags at a given position are
            allowed during training.

        Returns
        -------
        torch.FloatTensor
            Shape: ...
            The loss
        """
        transitions = self.transitions.masked_fill(
            self.forbidden_transitions, IMPOSSIBLE
        )
        start_transitions = end_transitions = torch.zeros_like(self.start_transitions)
        if self.with_start_end_transitions:
            start_transitions = self.start_transitions.masked_fill(
                self.start_forbidden_transitions, IMPOSSIBLE
            )
            end_transitions = self.end_transitions.masked_fill(
                self.end_forbidden_transitions, IMPOSSIBLE
            )

        # emissions: n_samples * n_tokens * ... * n_tags
        # bi_emissions: n_tokens * (2 * n_samples *  * ... * n_tags)
        bi_emissions = torch.stack(
            [emissions.masked_fill(~target, IMPOSSIBLE), emissions], 0
        ).unbind(2)

        out = [bi_emissions[0] + start_transitions]

        for word_bi_emissions in bi_emissions[1:]:
            res = logsumexp_reduce(out[-1], transitions)
            out.append(res + word_bi_emissions)

        last_out = (
            torch.stack(
                [
                    out[length - 1][:, i]
                    for i, length in enumerate(mask.long().sum(1).tolist())
                ],
                dim=1,
            )
            + end_transitions
        )

        supervised_z, unsupervised_z = last_out.logsumexp(-1)

        return -(supervised_z - unsupervised_z)


# noinspection PyTypeChecker
class MultiLabelBIOULDecoder(LinearChainCRF):
    def __init__(
        self,
        num_labels,
        with_start_end_transitions=True,
        learnable_transitions=True,
    ):
        """
        Create a linear chain CRF with hard constraints to enforce the BIOUL tagging
        scheme

        Parameters
        ----------
        num_labels: int
        with_start_end_transitions: bool
        learnable_transitions: bool
        """
        O, I, B, L, U = 0, 1, 2, 3, 4  # noqa: E741

        num_tags = 1 + num_labels * 4
        self.num_tags = num_tags
        forbidden_transitions = torch.ones(num_tags, num_tags, dtype=torch.bool)
        forbidden_transitions[O, O] = 0  # O to O
        for i in range(num_labels):
            STRIDE = 4 * i
            for j in range(num_labels):
                STRIDE_J = j * 4
                forbidden_transitions[L + STRIDE, B + STRIDE_J] = 0  # L-i to B-j
                forbidden_transitions[L + STRIDE, U + STRIDE_J] = 0  # L-i to U-j
                forbidden_transitions[U + STRIDE, B + STRIDE_J] = 0  # U-i to B-j
                forbidden_transitions[U + STRIDE, U + STRIDE_J] = 0  # U-i to U-j

            forbidden_transitions[O, B + STRIDE] = 0  # O to B-i
            forbidden_transitions[B + STRIDE, I + STRIDE] = 0  # B-i to I-i
            forbidden_transitions[I + STRIDE, I + STRIDE] = 0  # I-i to I-i
            forbidden_transitions[I + STRIDE, L + STRIDE] = 0  # I-i to L-i
            forbidden_transitions[B + STRIDE, L + STRIDE] = 0  # B-i to L-i

            forbidden_transitions[L + STRIDE, O] = 0  # L-i to O
            forbidden_transitions[O, U + STRIDE] = 0  # O to U-i
            forbidden_transitions[U + STRIDE, O] = 0  # U-i to O

        start_forbidden_transitions = torch.zeros(num_tags, dtype=torch.bool)
        for i in range(num_labels):
            STRIDE = 4 * i
            start_forbidden_transitions[I + STRIDE] = 1  # forbidden to start by I-i
            start_forbidden_transitions[L + STRIDE] = 1  # forbidden to start by L-i

        end_forbidden_transitions = torch.zeros(num_tags, dtype=torch.bool)
        for i in range(num_labels):
            STRIDE = 4 * i
            end_forbidden_transitions[I + STRIDE] = 1  # forbidden to end by I-i
            end_forbidden_transitions[B + STRIDE] = 1  # forbidden to end by B-i

        super().__init__(
            forbidden_transitions,
            start_forbidden_transitions,
            end_forbidden_transitions,
            with_start_end_transitions=with_start_end_transitions,
            learnable_transitions=learnable_transitions,
        )

    @staticmethod
    def tags_to_spans(tags):
        """
        Convert a sequence of multiple label BIOUL tags to a sequence of spans

        Parameters
        ----------
        tags: torch.LongTensor
            Shape: n_samples * n_tokens * n_labels

        Returns
        -------
        torch.LongTensor
            Shape: n_spans *  4
            (doc_idx, begin, end, label_idx)
        """

        # Note: tags are O, I, B, L, U => 0, 1, 2, 3, 4

        # begins_indices = torch.nonzero((tags == 4) | (tags == 2))
        # ends_indices = torch.nonzero((tags == 4) | (tags == 3))
        tags = tags.transpose(1, 2)

        tags_after = tags.roll(-1, 2)
        tags_before = tags.roll(1, 2)
        if 0 not in tags.shape:
            tags_after[..., -1] = 0
            tags_before[..., 0] = 0

        # A span starts if:
        # - tags is B / U
        # - tags is I / L and tags_before is not B / I (illegal transition)
        #   this gives: O -> I, L -> I, U -> I

        # A span ends if:
        # - tags is L / U
        # - tags is I / B and tags_after is not L / I (illegal transition)

        begins_indices = torch.nonzero(
            (tags == 4)
            | (tags == 2)
            | (((tags == 1) | (tags == 3)) & (tags_before != 2) & (tags_before != 1))
        )
        ends_indices = torch.nonzero(
            (tags == 4)
            | (tags == 3)
            | (((tags == 1) | (tags == 2)) & (tags_after != 3) & (tags_after != 1))
        )

        return torch.cat(
            [
                begins_indices[..., :3],
                ends_indices[..., [2]] + 1,
            ],
            dim=-1,
        )



from typing import Union

import spacy

regex = [r"\bsofa\b"]

digits = r"[^\d]*(\d*)"

value_extract = [
    dict(
        name="method_max",
        regex=r"(max)",
        reduce_mode="keep_first",
    ),
    dict(
        name="method_24h",
        regex=r"(24h)",
        reduce_mode="keep_first",
    ),
    dict(
        name="method_adm",
        regex=r"(admission)",
        reduce_mode="keep_first",
    ),
    dict(
        name="value",
        regex=r"^.*?[\n\W]*?(\d+)(?![h0-9])",
    ),
]

score_normalization_str = "score_normalization.sofa"


@spacy.registry.misc(score_normalization_str)
def score_normalization(extracted_score: Union[str, None]):
    """
    Sofa score normalization.
    If available, returns the integer value of the SOFA score.
    """
    score_range = list(range(0, 30))
    if (extracted_score is not None) and (int(extracted_score) in score_range):
        return int(extracted_score)



prefix_pattern = r"(?P<prefix>[cpPyraum]p?)"
tumour_pattern = r"T\s?(?P<tumour>([0-4o]|is))?(?P<tumour_specification>[abcdx]|mi)?"
tumour_pattern += r"(?:\((?P<tumour_suffix>[^()]{1,10})\))?"
node_pattern = r"(\s{,2}\/?\s{,2}([cpPyraum]p?)?\s{,2}N\s?(?P<node>[0-3o]|x)"
node_pattern += (
    r"(?P<node_specification>[abcdx]|mi)?(?:\((?P<node_suffix>[^()]{1,10})\))?)"
)

metastasis_pattern = (
    r"(\s{,2}\/?\s{,2}([cpPyraum]p?)?\s{,2}M\s?(?P<metastasis>([01o]|x))x?)"  # noqa: E501
)
resection_completeness = r"(\s{,2}\/?\s{,2}R\s?(?P<resection_completeness>[012]))"

version_pattern = (
    r"\(?(?P<version>uicc|accj|tnm|UICC|ACCJ|TNM)"
    r"\s+([éeE]ditions|[éeE]d\.?)?\s{,2}?"
    r"(?P<version_year>\d{4}|\d{2})\)?"
)

spacer = r"(.|\n){1,5}"

tnm_pattern = f"(?<={version_pattern}{spacer})?"
tnm_pattern += prefix_pattern + r"\s{,2}?" + f"({tumour_pattern})"
tnm_pattern += r"(\s{,2}" + f"{node_pattern})?"
tnm_pattern += r"(\s{,2}" + f"{metastasis_pattern})?"
tnm_pattern += r"(\s{,2}" + f"{resection_completeness})?"
tnm_pattern += f"({spacer}{version_pattern})?"
tnm_pattern = r"(?:\b|^)" + tnm_pattern + r"(?:\b|$)"



import pytest
from mlconjug3 import Conjugator

from edsnlp.conjugator import conjugate, conjugate_verb, get_conjugated_verbs

pytestmark = pytest.mark.filterwarnings("ignore")


def test_conjugate_verb():

    conjugator = Conjugator("fr")

    tests = [
        (("aimer", "Indicatif", "Présent", "1s"), "aime"),
        (("aimer", "Indicatif", "Présent", "2s"), "aimes"),
        (("aimer", "Indicatif", "Présent", "1p"), "aimons"),
        (("aimer", "Indicatif", "Présent", "2p"), "aimez"),
        (("aimer", "Indicatif", "Présent", "3p"), "aiment"),
    ]

    verb = "aimer"

    df = conjugate_verb(verb, conjugator=conjugator)

    for (v, m, t, p), term in tests:
        row = df.query("verb == @v & mode == @m & tense == @t & person == @p").iloc[0]
        assert row.term == term


def test_conjugate():

    tests = [
        (("aimer", "Indicatif", "Présent", "1s"), "aime"),
        (("aimer", "Indicatif", "Présent", "2s"), "aimes"),
        (("aimer", "Indicatif", "Présent", "1p"), "aimons"),
        (("aimer", "Indicatif", "Présent", "2p"), "aimez"),
        (("aimer", "Indicatif", "Présent", "3p"), "aiment"),
        (("convaincre", "Indicatif", "Présent", "2s"), "convaincs"),
        (("convaincre", "Indicatif", "Présent", "1s"), "convaincs"),
        (("convaincre", "Indicatif", "Présent", "1p"), "convainquons"),
        (("convaincre", "Indicatif", "Présent", "2p"), "convainquez"),
        (("convaincre", "Indicatif", "Présent", "3p"), "convainquent"),
    ]

    conjugate("aimer")
    df = conjugate(["aimer", "convaincre"])

    for (v, m, t, p), term in tests:
        row = df.query("verb == @v & mode == @m & tense == @t & person == @p").iloc[0]
        assert row.term == term


def test_get_conjugated_verbs():

    terms = get_conjugated_verbs(
        ["aimer", "convaincre"],
        matches=[dict(mode="Indicatif", tense="Présent")],
    )

    get_conjugated_verbs(
        "aimer",
        matches=dict(mode="Indicatif", tense="Présent"),
    )

    assert set(terms) == {
        "aime",
        "aimes",
        "aimons",
        "aimez",
        "aiment",
        "convainc",
        "convaincs",
        "convainquons",
        "convainquez",
        "convainquent",
    }
